#include <stdio.h>
#include <string.h>

// Command Line Arguments Definitions:

#define STARTCOMPILER	"--compiler"
#define STARTASSEMBLER	"--assembler"
#define STARTSQUISHER	"--squisher"
#define STARTBIN2HEX	"--bintohex"
#define COMPILEONLY	"-c"
#define ASSEMBLEONLY	"-a"
#define TASMOUTPUT	"--tasm"
#define VERBOSE		"-v"
#define OS		"--os"
#define EMULATOR	"--emulator"

#define SPACER		"---------------\n"


int compiler(int, char *[]);
int squisher(int, char *[]);
int bintohex(int, char *[]);
int assembler(int, char *[]);

int verbose;
int tasm;
int os;

int shiftargs(int, char *[], char **);
void errun(char *);

int main(int argc, char *argv[])
{
	int i;
	int args;
	int foundname = 0;
	char *sendargs[16];
	char fname[256];
	char inname[256];
	char calcname[10];
	char *infile;

	// Flags

	int compile = 1;
	int assemble = 1;
	int squish = 1;
	int mkos = 0;
	int emu = 0;
//	int tasm = 0;
//	int verbose = 0;
		
	tasm = 0;
	verbose = 0;

	sendargs[0] = argv[0];

	if (argc < 2) 
	{
		fprintf(stderr, "Syntax: bzc [ -c | -a | --tasm ] [ -v ] <input_file>\n");
		return (-1);
	}

	if (!strcmp(argv[1], STARTCOMPILER))
	{
		verbose = 1;
		args = shiftargs(argc, argv, sendargs);
		return(compiler(args, sendargs));
	}
	else if (!strcmp(argv[1], STARTASSEMBLER))
	{
		verbose = 1;
		args = shiftargs(argc, argv, sendargs);
		return(assembler(args, sendargs));
	}
	else if (!strcmp(argv[1], STARTSQUISHER))
	{
		verbose = 1;
		args = shiftargs(argc, argv, sendargs);
		return(squisher(args, sendargs));
	}
	else if (!strcmp(argv[1], STARTBIN2HEX))
	{
		verbose = 1;
		args = shiftargs(argc, argv, sendargs);
		return(bintohex(args, sendargs));
	}

	for(i = 1; i < argc; i++)
		if (!strcmp(argv[i], COMPILEONLY))
		{
			assemble = 0;
			squish = 0;
		}
		else if (!strcmp(argv[i], ASSEMBLEONLY))
			if (os)
				mkos = 0;
			else
				squish = 0;
		
		else if (!strcmp(argv[i], TASMOUTPUT))
		{
			assemble = 0;
			squish = 0;
			tasm = 1;
		}
		else if (!strcmp(argv[i], EMULATOR))
			emu = 1;
		else if (!strcmp(argv[i], OS))
		{	
			os = 1;
			if (squish)
			{
				mkos = 1;
				squish = 0;
			}
		}
		else if (!strcmp(argv[i], VERBOSE))
			verbose = 1;
		else if (!foundname)
			foundname = i;
		else
		{
			fprintf(stderr, "Syntax: bzc [ -c | -a | --tasm ] [ -v ] <input_file>\n");
			return(-1);
		}

	if (!foundname) 
	{
		fprintf(stderr, "Syntax: bzc [ -c | -a | --tasm ] [ -v ] <input_file>\n");
		return(-1);
	}

	strcpy(fname, argv[foundname]);
	infile = fname;

	for (i = 0; *infile++; i++);

	if (i < 5 || *(infile - 5) != '.') 
	{
		fprintf(stderr, "Incorrect Input File: \"%s\"\n", fname);
		return(-1);
	}

	infile -= 4;

	if (!strcmp(infile, "asm") || !strcmp(infile, "ASM"))
		compile = 0;
	else if (!strcmp(infile, "bin") || !strcmp(infile, "BIN"))
	{
		compile = 0;
		assemble = 0;
	}

	if (compile)
	{
		strcpy(inname, fname);
		*infile++ = 'a';
		*infile++ = 's';
		*infile++ = 'm';
		infile -= 3;

		sendargs[1] = inname;
		sendargs[2] = fname;
		args = 3;
		
		if (i = compiler(args, sendargs))
			return(i);
	}

	if (verbose) fputs(SPACER, stderr);

	if (assemble)
	{
		strcpy(inname, fname);
		*infile++ = 'b';
		*infile++ = 'i';
		*infile++ = 'n';
		infile -= 3;

		sendargs[1] = "-o";
		sendargs[2] = fname;
		sendargs[3] = inname;
		args = 4;
		
		i = assembler(args, sendargs);
		
		if (i)
		{
			fputs(SPACER, stderr);
			errun("assembler");
			return(i);
		}
	}

	if (verbose) fputs(SPACER, stderr);

	if (squish)
	{
		strcpy(inname, fname);

		*infile++ = '8';
		*infile++ = 'x';
		*infile++ = 'p';

		for (; *--infile != '/' && *infile != '\\' && infile != fname;);
		if (*infile == '/' || *infile =='\\') infile++;

		strncpy(calcname, infile, 8);
		calcname[9] = '\0';

		for(infile = calcname; *infile != '\0' && *infile != '.'; infile++);
		if (*infile == '.') *infile = '\0';

		sendargs[1] = "-i";
		sendargs[2] = inname;
		sendargs[3] = "-o";
		sendargs[4] = fname;
		sendargs[5] = "-n";
		sendargs[6] = calcname;
		sendargs[7] = "-u";
		sendargs[8] = "-83p";
		args = 9;
		
		i = squisher(args, sendargs);
		
		if (i)
		{
			fputs(SPACER, stderr);
			errun("squisher");
			return(i);
		}
	}
	else if (mkos)
	{
		*--infile = '\0';
		
		strcpy(inname, fname);

		*infile++ = '.';
		*infile++ = 'h';
		*infile++ = 'e';
		*infile++ = 'x';

		sendargs[1] = inname;
		args = 2;
		
		i = bintohex(args, sendargs);
		
		if (i)
		{
			fputs(SPACER, stderr);
			errun("to-hex-converter");
			return(i);
		}
	}

	return(0);
}

int shiftargs(int argc, char *argv[], char *sendargs[])
{
	int i;
	
	for (i = 2; i < argc; i++)
		sendargs[i-1] = argv[i];
	
	return (argc - 1);
}

void errun (char *mes)
{
	fprintf(stderr, "BZC-STARTER: Error while running the build in %s.\n", mes);
	fputs(	"BZC-STARTER: If you don't know the meaning of this\n"
		"BZC-STARTER: message, please write an email to:\n"
		"BZC-STARTER: bugs@hofhom.nl (with the subject \"BZC\")\n"
		"BZC-STARTER: and attach your source file.\n", stderr);
}

