/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "config.h"	// Load compile configuration.
#include "dep.h"	// Calculate configuration dependencies.

#define FNAME_LENGTH	96	// Maximum length of a filename.
#define LINE_LENGTH	128	// Maximum length of a line.

#define VAR_STRING_LENGTH	LINE_LENGTH	// Maximum length of a var
						// of the string type.

#define VAR_LENGTH 	VAR_STRING_LENGTH	// Maximum length of the name
						// of a VAR or FUNCTION.

#define MAX_SPRITE_LINES	128	// Maximum of lines / width (in bytes)
					// of a sprite. (for example: sprite 12x48,
					// = (2 bytes)x(48 lines) = 96

#define MAX_TERMS	32	// Maximum of keywords on a line.
#define MAX_FILES	64	// Maximum of source files (main source + includes).

#define CSMAX		64	// Maximal number of Control Structures inside another.

#define MAXPTRS		128	// Maximal number of both MEM's and SPRITE's.

#define CL_ARG_LENGTH	256	// Maximum length of a command line arg.

#define RESERVED_NUM	63	// The EXACT number of reserved KEYWORDS,
#define SYSTEM_AT	15	// The place of "SYSVAR" in the reserved KEYWORDS,
				// (see also bottom of this file).

/*********** Definitions for the getmem() function **********/

#define BLOCKSIZE	8192	// Bytes to malloc per block.
#define MAXBLOCKS	1024	// Used for a pointer array.

/********************** PATH definitions ********************/

#ifdef UNIX
#define HEADER_PATH	"/usr/share/bzc/include/"
#else
#define HEADER_PATH	"c:\\windows\\include\\bzc\\"
#endif

/************** The exit_fail function MACRO'S **************/

#define exit_nomem()			exit_fail(NoMem)
#define exit_io()			exit_fail(ioErr)
#define exit_maxblcks()			exit_fail(LowMaxBlocks)
#define exit_maxfiles()			exit_fail(LowMaxFiles)
#define exit_lowarr()			exit_fail(LowMaxPointers)
#define exit_lowcl()			exit_fail(LowCL)
#define exit_LowSprite()		exit_fail(LowSpriteLines)
#define exit_lowfname()			exit_fail(LowFnameLength)
#define exit_lowline()			exit_fail(LowLine)
#define exit_inc()			exit_fail(WrongInc)
#define exit_nomath()			exit_fail(NoMath)
#define exit_lowcsmax()			exit_fail(LowCS)


/**************** The clmes function MACRO'S ****************/

#define clmes_old(x)			clmes(x, file, 0, 0, 0, NULL)
#define clmes_line(x, y)		clmes(y, fline, x, 0, 0, NULL)
#define clmes_syntax(x)			clmes(x, flsyn, line, 0, 0, NULL)
#define clmes_syntax_dec(x, y)		clmes(x, flsyn_dec, line, 0, y, NULL)
#define clmes_syntax_keyw(x)		clmes(x, flsyn_keyw, line, 0, 0, NULL)
#define clmes_notdecl(x)		clmes(x, flsyn_notdec, line, 0, 0, NULL)
#define clmes_mem_dp(x, y, z) 		clmes(x, memdebug_dp, 0, y, 0, z)
#define clmes_mem_cdpd(a, b, c, d)	clmes(a, memdebug_cdpd, 0, b, d, c)
#define clmes_debug(x, y)		clmes(x, debug, y, 0, 0, NULL)
#define clmes_debug_s(x, y, z)		clmes(x, debug_s, z, 0, 0, y)
#define clmes_debug_ds(a, b, c, d)	clmes(a, debug_ds, d, b, 0, c)
#define clmes_olddebug(x)		clmes_debug(x, 0)

/************** The codedoline function MACRO'S *************/

#define codeline(x)			codedoline(x, 0)
#define codelastline(x)			codedoline(x, 1)

/***************** The VAR TYPE "enumeration" ***************/

#define VAR_SYSTEM	1
#define VAR_STRING	2
#define VAR_BYTE	3
#define VAR_INT		4
#define VAR_CONST	5

/***************** The ERROR CODE enumeration ***************/

enum ExitCodes
{
	NoMem = 1,
	ioErr,
	LowMaxBlocks,
	LowMaxFiles,
	LowMaxPointers,
	WrongInc,
	NoMath,
	LowCL,
	LowSpriteLines,
	LowFnameLength,
	LowLine,
	LowCS,
	SynErr = 15
};

/**************** The MESSAGE TYPE enumeration **************/

enum MesType
{
	file,
	fline,
	flsyn,
	flsyn_dec,
	flsyn_keyw,
	flsyn_notdec,
	memdebug_dp,
	memdebug_cdpd,
	debug,
	debug_s,
	debug_ds
};

/************** The TYPE DECLARATION enumeration ************/

enum SaveType
{
	OldType = -1,
	reg8 = 0,
	reg16,
	var,
	function_user,
	function_rcall,
	romcall,
	array
};

/*********** The Control structure type enumeration *********/

enum CtrlType
{
	Repeat,
	ForNext,
	Loop,
	IfThenElse
};

/*************** The COMPILING FLAG enumeration *************/

enum CompFlags
{
	Noflag,
	Start,
	Code,
	CompCode,
	Asm
};

/**************** The includes table struct *****************/

struct Includes
{
	char *File;			// Pointer to the file name.
	struct Includes *Next;
};

/************** The structs needed for file io **************/

struct FileLine
{
	int FileName;			// Index for Files[] array.
	int Line;			// Line number.
	int NumTerms;			// Number of keywords on this line.
	char *Terms[MAX_TERMS + 1];	// A pointer to all the keywords.
	char Quoted[MAX_TERMS + 1];	// Is the keyword quoted?
	struct FileLine *Next;		// A pointer to the next line.
};

/*************** The var/function table struct **************/

struct SaveTab
{
	int Line;			// * The line where the CODE (or ASM) keyword is.
	int Lastline;			// * The line where the ENDCODE (or ENDASM) keyword
					// is OR which kind of var this is.
	int Used;			// How many times is this function or var used?
	int Args;			// * How many function arguments are there? OR
					//   the value of a var.
	char Name[VAR_LENGTH];		// The name of the function / var.
	char Realname[VAR_LENGTH];	// * The name of the romcall or system var.
	enum SaveType Type;		// Is it a function or romcall or var or ???
	struct SaveTab *Next;		// Pointer to the next table entry.
//	int Tabnum; 			Not yet available, the table entry number.
};
				// * means optional (not always in use).

/************* The struct for control structures *************/

struct Control
{
	enum CtrlType Type;
	struct SaveTab *Var;
	char *ID;
	int Special;
};

/********************* GLOBAL VARIABLES *********************/

EXTERN int line;			// Line counter while compiling.
EXTERN int cmperr;			// Error code while compiling.

EXTERN char *progname;			// Used for better messaging, check clmes().

EXTERN enum CompFlags comflag;		// Whether we are in a function, the main
EXTERN int AsmCom;			// function, an ASM part or in none.
EXTERN struct SaveTab *CurFunc;		// Current compiling function.

EXTERN struct SaveTab *tabptr;		// Pointer to the first table (used to store vars
					// functions, romcalls, etc.

EXTERN char *Files[MAX_FILES];		// A source filename array.
EXTERN struct FileLine *flindex;	// A pointer to the first source line.

EXTERN struct Includes *incindx;	// A pointer to the first include file 
					// (NEW INCLUDE * not #include *)

EXTERN struct Control cstructs[CSMAX];	// A Control Structures Array;
EXTERN int curcstruct;			// With his index.

EXTERN void *memptr[MAXBLOCKS];		// Pointer array for getmem().
EXTERN unsigned long memloc;		// 'cursor' inside getmem().

EXTERN int *ptrs[MAXPTRS];		// A pointers array, speccially for ARRAY's.
EXTERN int curptr;			// The last record of *ptrs[]

extern int verbose;
extern int tasm;
extern int os;

EXTERN int idnum;
EXTERN char *curfunc;
EXTERN int globalvar;
EXTERN int pushes;

/******************* FUNCTIONS IN main.c *******************/

void exit_fail(enum ExitCodes);		// Exit if a system error appears.
void transtoup(char *, char *);		// Uppercase a string.
void clmes(char *, enum MesType, \
		int, int, int, void *);	// Command Line MESsage.
int compile(FILE *, FILE *);		// The 'main' program, calls all other functions.
int poweroftwo(int);			// returns 2 to the power of x.

/******************* FUNCTIONS IN file.c *******************/

void header(FILE *);			// Write the Z80 (T)ASM header to the file.
int footer(FILE *);			// Write the Z80 (T)ASM footer to the file 
					// and close it.
int writefuncs(FILE *);		// Write the Z80 functions to the file
void writevars(FILE *);			// Write the Z80 vars to the file.
struct FileLine *load(FILE *, \
		struct FileLine *);	// Put the source and incudes as one file in the 
					// memory.

/******************** FUNCTIONS IN mem.c ********************/

void initmem(void);				// Put's zero's in several global vars.
void freemem(void);				// Free all memory.

struct FileLine *gotoline(int);			// Gives a pointer to the (int) source line.
void wholeline(char *, struct FileLine *);	// Makes a string from a FileLine pointer.
void term(struct FileLine *, char *, int);	// Put flptr->Terms[int] into [char *].
void addinc(char *);				// Add an include file to the include file table.
void *getmem(int);				// A malloc()-like function, so we don't have to
						// malloc() every 6 bytes.
int ptrtoint(int *);				// "converts" an pointer to an int (for ARRAY's).

#ifdef MEMORY_DUMP_ON_EXIT_FAILURE

void dumpmem(void);				// dump all the getmem()-memory to "memory.dmp"

#endif

/******************* FUNCTIONS IN line.c ********************/

int equterm(struct FileLine *, char *, int);	// strcmp() a term (and check if it is empty)
int todigit(char *);				// Converts a string to an int.
int checkex(char *, struct FileLine *);		// Checks if *str already exists in the SaveTab
						// tables and if it is a keyword.
int examline(FILE *);				// Check command of every line and
						// call 'his' function.
void examterm(struct FileLine *, FILE *);	// Same but for single terms.
char *mkasm(char *);				// Makes a line good for ASM.

/******************* FUNCTIONS IN table.c *******************/

void init_reg(void);			// Load the Z80 registers to the table.
struct SaveTab *totable(enum SaveType, char *, \
	char *, int , int, int);	// Add an entry to the table.
int usetable(char *);			// Increases the Used entry of the given table.
//int codedoline(char *, int);		// Sets the line or lastline entry of the given table to line.
int ffirstline(char *);			// Sets the line entry of the given function to line.
void flastline(void);			// Same for lastline.

struct SaveTab *addtable(void);		// Create a new entry.
struct SaveTab *findtable(char *, int);	// Gives the pointer to the table with given name.
void forcevar(struct FileLine *);	// Forces a var to be used.
char *makeid(char *, char *);		// Create an VAR or Function identity.

/****************** FUNCTIONS IN compile.c ******************/

	// FUNCTION:					// CODE:

//void anal_new(struct FileLine *, FILE *);	//	NEW *
void anal_code(struct FileLine *, FILE *);	//	CODE
void anal_endcode(struct FileLine *, FILE *);	//	ENDCODE
void anal_asm(struct FileLine *, FILE *);	//	ASM
void anal_endasm(struct FileLine *, FILE *);	//	ENDASM
void anal_start(struct FileLine *, FILE *);	//	START
void anal_end(struct FileLine *, FILE *);	//	END
void anal_exit(struct FileLine *, FILE *);	//	EXIT
void anal_return(struct FileLine *, FILE *);	//	RETURN
void anal_printf(struct FileLine *, FILE *);	//	PRINTF
void anal_for(struct FileLine *, FILE *);	//	FOR *
void anal_next(struct FileLine *, FILE *);	//	NEXT
void anal_repeat(struct FileLine *, FILE *);	//	REPEAT *
void anal_endrepeat(struct FileLine *, FILE *);	//	ENDREPEAT
void anal_if(struct FileLine *, FILE *);	//	IF *
void anal_else(struct FileLine *, FILE *);	//	ELSE
void anal_endif(struct FileLine *, FILE *);	//	ENDIF
void anal_loop(struct FileLine *, FILE *);	//	LOOP [ WHILE | UNTIL ] *
void anal_endloop(struct FileLine *, FILE *);	//	ENDLOOP


int anal_varfunc(struct FileLine *, FILE *);	//	[ Function ] * || [ Variable] *

int solve(int, int, struct FileLine, FILE*);	//	Calculations.
int solvetest(int, int, struct FileLine, int, FILE*);	//	Logical Tests.
int ifoperator(char *);				//  Test if a string is a
						// logical operator.

/******************** FUNCTIONS IN new.c ********************/

//void newinc(struct FileLine *);	//      INCLUDE * -> done in mem.c
void newreg8(struct FileLine *);	//      REG8 *
void newreg16(struct FileLine *);	//      REG16 *
void newfunc(struct FileLine *);	//      FUNCTION *
void newrcall(struct FileLine *,FILE *);//      ROMCALL *
void newvar(struct FileLine *, FILE *);	//      [ INT | BYTE | ... ]
void newsprite(struct FileLine *);	//	SPRITE *
void newarray(struct FileLine *);	//	MEM *

void addand(char *, char *);		//	Add a & before a varname.

/****************** FUNCTIONS in function.c *****************/

int compilefunction(struct SaveTab *, FILE *);
int pushvar(struct FileLine *, FILE *, int);
int popvar(char *, FILE *);
int toHL(struct SaveTab *, FILE *, int);
char *getarg(struct SaveTab *, int);
int toHLpp(struct FileLine, int, int, FILE*, int);

/******************** The KEYWORDS table ********************/

#ifndef DECL_VARS

extern char *KeyWords[RESERVED_NUM];

#else

char *KeyWords[RESERVED_NUM] = {
	
	"NEW",		// 6	-	8
	"REG8",
	"REG16",
	"ROMCALL",
	"FUNCTION",
	"VAR",

	"SPRITE",	// 2
	"FROM",

	"MEM",		// 1	-	15
	
	"INCLUDE",	// 3
	"LIB",
	"ROUTINE",

	"ARG",		// 2
	"USER",

	"SYSVAR",	// 5
	"STRING",
	"BYTE",
	"INT",
	"CONST",
	

	"START",	// 2
	"END",

	"FSTART",	// 2
	"FEND",

	"ASM",		// 2	-	40
	"ENDASM",

	"=",		// 5
	"==",
	"!=",
	"+",
	"+=",
	"-",		// 5
	"-=",
	"*",
	"*=",
	"/",
	"/=",		// 5
	"%",
	"%=",
	"<",
	">",
	"<=",		// 5
	">=",
	"(",
	")",
	"&&",
	"||",		// 3
	"AND",
	"OF",
	
	"IF",		// 4
	"THEN",
	"ELSE",
	"ENDIF",

	"TO",		// 4
	"FOR",
	"NEXT",
	"BACK",

	"LOOP",		// 4
	"ENDLOOP",
	"WHILE",
	"UNTIL",

	"REPEAT",	// 2
	"ENDREPEAT",

	"FORCE"		// 1
};

#endif
