/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>

#include "bzc.h"

void anal_asm(struct FileLine *flptr, FILE *file)
{
#ifdef CODE_DEBUG

	clmes_debug("ASM keyword found.", 1);
	
#endif
	if (flptr->NumTerms != 1)
		clmes_syntax("No RETURN after ASM.");
	else
		AsmCom = 1;
}

void anal_code(struct FileLine *flptr, FILE *file)
{
	int result;
	char *mes = "Expecting FSTART <function_name> <arguments>,";
	
#ifdef CODE_DEBUG

	clmes_debug("CODE keyword found.", 1);
	
#endif
	if (flptr->NumTerms == 1)
	{
		clmes_syntax(mes);
		clmes_syntax("but return at <function_name>.");
		return;
	}
	
	result = ffirstline(flptr->Terms[2]);
	
	switch (result)
	{
		case 1:
			clmes_syntax(mes);
			clmes_syntax("but <function_name> is not declared as function.");
			return;
		case 2:
			clmes_syntax(mes);
			clmes_syntax("but <function_name> is not a function.");
			return;
	}
	
	if (CurFunc->Args != flptr->NumTerms - 2)
	{
		clmes_syntax("Number of arguments declared for this function differs");
		clmes_syntax("from number of used arguments.");
	}
}

void anal_endcode(struct FileLine *flptr, FILE *file)
{
#ifdef CODE_DEBUG
	
	clmes_debug("ENDCODE keyword found.", 1);
#endif
	comflag = Noflag;
}

void anal_start(struct FileLine *flptr, FILE *file)
{
#ifdef START_DEBUG
	
	clmes_debug("START keyword found.", 1);
#endif
	comflag = Start;
	curfunc = "main";
}

void anal_end(struct FileLine *flptr, FILE *file)
{
#ifdef START_DEBUG
	
	clmes_debug("END keyword found.", 1);
#endif
	comflag = Noflag;
	curfunc = "";
}

void anal_printf(struct FileLine *flptr, FILE *file)
{
	int i = 0;
	struct SaveTab *temptab;

	if (flptr->NumTerms != 2 && flptr->NumTerms != 4)
	{
		clmes_syntax("Expecting PRINTF [<row> <column>] <var>.");
		return;
	}
	if (!findtable("locate", 0) || !findtable("printf", 0) ||
		!findtable("str", 0) || !findtable("num", 0))
	{
		clmes_syntax("I/O routines not loaded, please put");
		clmes_syntax("#include <io.h> or #include <asmlib.h>");
		clmes_syntax("on top of your source file.");
		return;
	}
	if (flptr->NumTerms == 4)
	{
		if (pushvar(flptr, file, 2) || pushvar(flptr, file, 3))
			return;
		fputs("    CALL locate\n", file);
		if ((temptab = findtable("locate", 0))->Used != -1) temptab->Used++;
	}
	
	if (!(temptab = findtable(flptr->Terms[flptr->NumTerms], 0)))
		fputs("    LD HL, str\n", file);
	else
	{
		if (temptab->Used != -1) temptab->Used++;

		if	(temptab->Type == var && temptab->Lastline == 2)
			fputs("    LD HL, str\n", file);
		else
			fputs("    LD HL, num\n", file);
	}
	fputs("    PUSH HL\n", file);

	if (pushvar(flptr, file, flptr->NumTerms))
		return;

	fputs("    CALL printf\n", file);

	if ((temptab = findtable("printf", 0))->Used != -1) temptab->Used++;
}

void anal_exit(struct FileLine *flptr, FILE *file)
{
	if (flptr->NumTerms != 1)
	{
		clmes_syntax("No arguments expected after EXIT.");
		return;
	}
	fputs("    LD SP, (__RETPointer)\n", file);
	fputs("    RET\n", file);
}

void anal_return(struct FileLine *flptr, FILE *file)
{
	int i;

	if (pushes)
		for (i = pushes; i; i--)
			fputs("    POP HL\n", file);

	if (flptr->NumTerms != 1)
		solve(2, flptr->NumTerms, *flptr, file);
	fputs("    RET\n", file);
}

int anal_varfunc(struct FileLine *flptr, FILE *file)
{
	int x;
	struct SaveTab *tabptr;
	struct SaveTab *functab;

	enum SaveType stype;
	int i = 1;

	if (!(tabptr = findtable(flptr->Terms[1], 0)))
		return(1);

	stype = tabptr->Type;

	if (stype == reg8 || stype == reg16 || stype == var)
	{
#ifdef VAR_DEBUG

		clmes_debug("Variable found.", 2);
#endif
//		toHL(tabptr, file, 0);
//		fputs("    EX DE, HL\n", file);
		
		switch (equterm(flptr, "=", 2))
		{
			case 0:
				if (flptr->NumTerms < 3)
				{
					clmes_syntax("Expecting <variable_name> <operator> <expression>,");
					clmes_syntax("but RETURN after <operator>.");
					return(-1);
				}
				x = solve(3, flptr->NumTerms, *flptr, file);
				if (x) return (2);
				break;
			case 1:
				if (flptr->NumTerms < 3)
				{
					clmes_syntax("Expecting <variable_name> <operator> <expression>,");
					clmes_syntax("but RETURN after <operator>.");
					return(-1);
				}

				x = solve(3, flptr->NumTerms, *flptr, file);
				if (x) return (2);

				fputs("    EX DE, HL\n", file);
				toHL(tabptr, file, 0);

				if (!equterm(flptr, "+=", 2))
					fputs("    ADD HL, DE\n", file);
				else if (!equterm(flptr, "-=", 2))
				{
					fputs("    SCF\n", file);
					fputs("    CCF\n", file);
					fputs("    SBC HL, DE\n", file);
				}
				else if (!equterm(flptr, "*=", 2))
					if (functab = findtable("__multiply__", 0))
					{
						fputs("    EX DE, HL\n", file);
						fputs("    CALL __multiply__\n", file);
						functab->Used++;
					}
					else
						exit_nomath();
				else if (!equterm(flptr, "/=", 2))
					if (functab = findtable("__devide__", 0))
					{
						fputs("    EX DE, HL\n", file);
						fputs("    CALL __devide__\n", file);
						functab->Used++;
					}
					else
						exit_nomath();
				else if (!equterm(flptr, "%=", 2))
					if (functab = findtable("__remainder__", 0))
					{
						fputs("    EX DE, HL\n", file);
						fputs("    CALL __remainder__\n", file);
						functab->Used++;
					}
					else
						exit_nomath();
				else if (equterm(flptr, "=", 2))
					clmes_syntax("Keyword 2 (after variable) is not an operator");
				break;
			case -1:
				clmes_syntax("Expecting Operator after variable.");
		}
		
		// Put HL into de Destination.
		toHL(tabptr, file, 1);	
		
		return(0);	
	}
	else if (stype == function_user || stype == function_rcall || stype == romcall)
	{
#ifdef FUNC_DEBUG

		clmes_debug("Function found.", 2);
#endif
		if (stype == romcall)
		{
			if (flptr->NumTerms - 1)
			{
				clmes_syntax("Arguments after a ROMCALL.");
				return(4);
			}
		
			fputs("    RST $28\n", file);
			fprintf(file, ".DW %s\n", *tabptr->Realname ? tabptr->Realname : tabptr->Name);
		}
		else if (stype == function_rcall && flptr->NumTerms - 1 != tabptr->Args)
		{
			clmes_syntax("Number of Used Arguments doensn't match the");
			clmes_syntax("declared number of arguments.");
			return(4);
		}
		else
		{
			// This slows down a lot for romcalls, rewrite in future.

			if (stype == function_rcall)
			{
				while (i < flptr->NumTerms)
					if (pushvar(flptr, file, ++i))
					{
						clmes_notdecl(flptr->Terms[i]);
						return(2);
					}
			
				for (i = flptr->NumTerms - 1; i; i--)
				
					if (popvar(getarg(tabptr, i), file))
					{
						line = tabptr->Line;
						clmes_notdecl(getarg(tabptr, i));
						return(3);
					}
				
				fputs("    RST $28\n", file);
				fprintf(file, ".DW %s\n", tabptr->Realname);
			}
			else	// User defined function.
			{
				// Call the function.
				// Now doing it by Solve()
			
				if (flptr->NumTerms == 1)
					fprintf(file, "    CALL %s\n", tabptr->Name);
				else				
					return(solve(1, flptr->NumTerms, *flptr, file));
			}
		}
		// Increase Used Flag.
	
		if (tabptr->Used != -1) tabptr->Used++;
	}
	else
		return(1);

	return(0);
}

int solve(int start, int end, struct FileLine fl, FILE *file)
{
	char c;
	char *exp;
	int i, j;
	int stored = 0;
	int numterms, numbrackets, infunc = 0;
	int TermStart[MAX_TERMS], TermType[MAX_TERMS];
	struct SaveTab *tabptr;
	struct SaveTab *multab;
	struct SaveTab *devtab;
	struct SaveTab *remtab;

	numterms = -1;
	numbrackets = 0;

	for (i = start; i <= end; i++)
	
		if (fl.Quoted[i])
		{
			TermType[++numterms] = 4;
			TermStart[numterms] = i;
		}
		else if (!strcmp(fl.Terms[i], "("))
		{
			if (!numbrackets++ && !infunc)
			{
				TermStart[++numterms] = i;
				j = i;
			}
		}
		else if (!strcmp(fl.Terms[i], ")"))
		{
			if (--numbrackets < 0)
			{
				clmes_syntax("More \")\"'s than \"(\"'s in expression.");
				return(-4);
			}
			if (!numbrackets && !infunc)
				TermType[j] = i - j;
		}
		else if (	// Ignore Operators
			!strcmp(fl.Terms[i], "+") ||
			!strcmp(fl.Terms[i], "-") ||
			!strcmp(fl.Terms[i], "*") ||
			!strcmp(fl.Terms[i], "/") ||
			!strcmp(fl.Terms[i], "%"))
		{
				if (i == end)
				{
					clmes_syntax("Expecting expression as last term in calculation.");
					return(-14);
				}
				else if (!numbrackets)
					infunc = 0;
		}
		else if (!numbrackets) // It's for us to solve.
		{
			if (todigit(fl.Terms[i]) != -1)
				TermType[++numterms] = -1;
			else if (!(tabptr = findtable(fl.Terms[i],0)))
			{
				clmes_syntax("Undeclared variable in expression.");
				clmes_notdecl(fl.Terms[i]);
				return(-1);
			}
			else	// Check Type.
			{
				if (tabptr->Used != -1) tabptr->Used++;

				if (tabptr->Type == romcall || tabptr->Type == function_rcall)
				{
					clmes_syntax("Romcall used as expression.");
					return(-2);
				}
				else if (tabptr->Type == function_user)
					if (infunc)
					{
						clmes_syntax("Function used as function argument.");
						clmes_syntax("Try using: <function> ( <function> <arg1> ... ) ...");
						return(-3);
					}
					else
					{
						infunc = 1;
						TermType[++numterms] = -2;	// Because infunc != 1, 
						TermStart[numterms++] = i;	// we need to add 1 'extra'
					}
				else
					TermType[++numterms] = -3;
			}
			if (infunc)
				numterms--;
			else
				TermStart[numterms] = i;
		}
	// Check term-syntax
	if (numbrackets)
	{
		clmes_syntax("More \"(\"'s than \")\"'s in expression.");
		return(-4);
	}
	if (numterms == -1)
	{
		clmes_syntax("No expressions.");
		return(-5);
	}
	if (TermStart[0] != start)
	{
		clmes_syntax("No expression before first operator.");
		return(-6);
	}
	if (!numterms)
	{
		if (toHLpp(fl, start, TermType[0], file, 0))
			return(-7);
	}
	else
	{
		for (i = 0; i < numterms; i++)
		{
			j = TermStart[i + 1] - TermStart[i] - 2;
			if (j)
				if (j < 0)
				{
					clmes_syntax("Expecting operator between expressions.");
					return(-8);
				}
				else if (TermType[i] > 0)
					if (j < TermType[i])
					{
						clmes_syntax("Expecting operator between expressions.");
						return(-9);
					}
					else
					{
						clmes_syntax("Expecting expression between operators,");
						return(-10);
					}
				else if (TermType[i] == -2)
				{
					tabptr = findtable(fl.Terms[TermStart[i]], 0);

					// Better run the whole test.
					//if (j != tabptr->Args)
						if (j < tabptr->Args)
						{
							clmes_syntax("Not enough function arguments.");
							return(-11);
						}
						else
						{
							infunc = 0;
							for (j = TermStart[i]+1; j < TermStart[i+1]-1; j++)
							{
								if(!strcmp(fl.Terms[j], "("))
								{
									numbrackets++;
									infunc++;
								}
								else if (!strcmp(fl.Terms[j], ")"))
								{
									numbrackets--;
								}
								else if (!numbrackets)
									infunc++;
							}
							if (infunc > tabptr->Args)
							{
								clmes_syntax("Too many function arguments.");
								return(-12);
							}
							else if (infunc < tabptr->Args)
							{
								clmes_syntax("Too few function arguments.");
								return(-12);
							}
						}
				}
				else
				{
					clmes_syntax("Expecting operator between expressions.");
					return(-13);
				}
		}

	// Begin calculating...

	/* Algorhitm:
	   * if the operator is * or / or %: calculate directly
	   * if the operator is - or + and stored == 1: calculate previous calculation and set stored to 0
	   * if the operator is - or + and stored == 0:
		* if the next operator is * or / or %: set stored = this position, calculate next.
		* else: calculate directly.
	*/

		if (toHLpp(fl, TermStart[0], TermType[0], file, 1)) return(-30);

		if (!((multab = findtable("__multiply__", 0)) && 
			(devtab = findtable("__devide__", 0)) &&
			(remtab = findtable("__remainder__", 0))))
			
			exit_nomath();

		for (i = 0; i < numterms;)
		{
			switch (c = *fl.Terms[TermStart[++i]-1])
			{
				case '*':
					fputs("    EX DE, HL\n", file);
					if (toHLpp(fl, TermStart[i], TermType[i], file, 1)) return(-30);
					fputs("    CALL __multiply__\n", file);
					if (multab->Used != -1) multab->Used++;
					break;
				case '/':
					fputs("    EX DE, HL\n", file);
					if (toHLpp(fl, TermStart[i], TermType[i], file, 1)) return(-30);
					fputs("    CALL __devide__\n", file);
					if (devtab->Used != -1) devtab->Used++;
					break;
				case '%':
					fputs("    EX DE, HL\n", file);
					if (toHLpp(fl, TermStart[i], TermType[i], file, 1)) return(-30);
					fputs("    CALL __remainder__\n", file);
					if (remtab->Used != -1) remtab->Used++;
					break;
				case '+':
				case '-':
					if (stored)
					{
						fputs("    POP DE\n", file);
						c = *fl.Terms[TermStart[stored]-1];
						stored = 0;

						if (c == '+')
							fputs("    ADD HL, DE\n", file);
						else
						{
							fputs("    EX DE, HL\n", file);
							fputs("    SCF\n", file);
							fputs("    CCF\n", file);
							fputs("    SBC HL, DE\n", file);
						}
					}
					if (i == numterms || *fl.Terms[TermStart[i+1]-1] == '+' ||
						*fl.Terms[TermStart[i+1]-1] == '-')
					{
						fputs("    EX DE, HL\n", file);
						exp = NULL;
						if (TermType[i] == -1)
							exp = fl.Terms[TermStart[i]];
						else if (TermType[i] == -4)
						{
							exp = makeid("__STRING","");
							globalvar = 1;
							tabptr = totable(var, exp, "", TermStart[i], line, 2);
							tabptr->Used = 1;
						}
						else if ((tabptr = findtable(fl.Terms[TermStart[i]], 0)) && tabptr->Type == var)
						{
							if (tabptr->Lastline == 2 || tabptr->Lastline == 5)
							{
								exp = *tabptr->Realname ? tabptr->Realname : tabptr->Name;
								if (tabptr->Used != -1) tabptr->Used++;
							}
						}
						if (toHLpp(fl, TermStart[i], TermType[i], file, 1)) return(-30);
					}
					else
					{
						stored = i;
						fputs("    PUSH HL\n", file);
						if (toHLpp(fl, TermStart[i], TermType[i], file, 1)) return(-30);
						break;
					}
					if (c == '+')
						fputs("    ADD HL, DE\n", file);
					else
					{
						fputs("    EX DE, HL\n", file);
						fputs("    SCF\n", file);
						fputs("    CCF\n", file);
						fputs("    SBC HL, DE\n", file);
					}
					
					break;
			}
		}
		if (stored)
		{
			fputs("    POP DE\n", file);

			if (*fl.Terms[TermStart[stored]-1] == '+')
				fputs("    ADD HL, DE\n", file);
			else
			{
				fputs("    EX DE, HL\n", file);
				fputs("    SCF\n", file);
				fputs("    CCF\n", file);
				fputs("    SBC HL, DE\n", file);
			}
		}
	}
	return(0);
}

void anal_for(struct FileLine *flptr, FILE *file)
{
	int i, found = 0;
	
	if (++curcstruct == CSMAX)
		exit_lowcsmax();
	if (strcmp(flptr->Terms[3], "=") || !(cstructs[curcstruct].Var = findtable(flptr->Terms[2], 0)))
	{
		clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
		return;
	}

	if (cstructs[curcstruct].Var->Used != -1)
		cstructs[curcstruct].Var->Used++;

	if (!strcmp(flptr->Terms[flptr->NumTerms], "BACK"))
		cstructs[curcstruct].Special = 1;
	else
		cstructs[curcstruct].Special = 0;

	if (flptr->NumTerms - cstructs[curcstruct].Special < 6)
	{
		clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
		return;
	}
	if (flptr->NumTerms - cstructs[curcstruct].Special == 6)
		if (strcmp(flptr->Terms[5], "TO"))
		{
			clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
			return;
		}
	if (!strcmp(flptr->Terms[4], "TO") || 
		!strcmp(flptr->Terms[flptr->NumTerms - cstructs[curcstruct].Special], "TO"))
	{
		clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
		return;
	}
	for (i=5; i < flptr->NumTerms - cstructs[curcstruct].Special; i++)
		if (!strcmp(flptr->Terms[i], "TO"))
			if (found)
			{
				clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
				return;
			}
			else
				found = i;
	if (!found)
	{
		clmes_syntax("Expecting FOR <var> = <expression> TO <expression> [ BACK ]");
		return;
	}
	if (solve(4, found - 1, *flptr, file))
		return;
	if (toHL(cstructs[curcstruct].Var, file, 1))
		return;
	if (solve(found + 1, flptr->NumTerms - cstructs[curcstruct].Special, *flptr, file))
		return;
	if (cstructs[curcstruct].Special)
		fputs("    DEC HL\n", file);
	else
		fputs("    INC HL\n", file);
	fputs("    PUSH HL\n", file);
	pushes++;

	cstructs[curcstruct].Type = ForNext;
	cstructs[curcstruct].ID = makeid("__FORNEXT","");
	fprintf(file, "%sBEGIN__:\n", cstructs[curcstruct].ID);
}
void anal_next(struct FileLine *flptr, FILE *file)
{
	if (curcstruct == -1 || cstructs[curcstruct].Type != ForNext)
	{
		clmes_syntax("NEXT without FOR");
		clmes_syntax("(or inside an other control structure).");
		return;
	}
	
	if (flptr->NumTerms != 1)
		if (flptr->NumTerms != 2 || findtable(flptr->Terms[2], 0) != cstructs[curcstruct].Var)
		{
			clmes_syntax("Expecting NEXT (without any arguments).");
			return;
		}
	
	if(toHL(cstructs[curcstruct].Var, file, 0))
		return;

	if (cstructs[curcstruct].Special)
		fputs("    DEC HL\n", file);
	else
		fputs("    INC HL\n", file);

	if(toHL(cstructs[curcstruct].Var, file, 1))
		return;

	fprintf(file,
		"    POP DE\n"
		"    DEC SP\n"
		"    DEC SP\n"
		"    LD A, D\n"
		"    CP H\n"
		"    JP NZ, %sBEGIN__\n"
		"    LD A, E\n"
		"    CP L\n"
		"    JP NZ, %sBEGIN__\n"
		"    POP HL\n",
		cstructs[curcstruct].ID,
		cstructs[curcstruct].ID);

	pushes--;
	curcstruct--;
}

void anal_repeat(struct FileLine *flptr, FILE *file)
{
	if (++curcstruct == CSMAX)
		exit_lowcsmax();
	if (flptr->NumTerms == 1)
	{
		clmes_syntax("Expecting REPEAT <times_to_repeat>.");
		return;
	}
	if (solve(2, flptr->NumTerms, *flptr, file))
		return;

	cstructs[curcstruct].Type = Repeat;
	cstructs[curcstruct].ID = makeid("__REPEAT","");

	pushes++;
	fputs("    PUSH HL\n", file);
	fprintf(file, "%sBEGIN__:\n", cstructs[curcstruct].ID);
	fputs(	"    POP BC\n"
		"    LD A, B\n"
		"    OR C\n", file);
	fprintf(file, "    JP Z, %sEND__\n", cstructs[curcstruct].ID);
	fputs(	"    DEC BC\n"
		"    PUSH BC\n", file);
}

void anal_endrepeat(struct FileLine *flptr, FILE *file)
{
	if (curcstruct == -1 || cstructs[curcstruct].Type != Repeat)
	{
		clmes_syntax("ENDREPEAT without REPEAT");
		clmes_syntax("(or inside an other control structure).");
		return;
	}
	if (flptr->NumTerms != 1)
	{
		clmes_syntax("Expecting <RETURN> after ENDREPEAT.");
		return;
	}
	pushes--;
	fprintf(file,
		"    JP %sBEGIN__\n"
		"%sEND__:\n",
		cstructs[curcstruct].ID, cstructs[curcstruct].ID);
	curcstruct--;
}
void anal_if(struct FileLine *flptr, FILE *file)
{
	if (++curcstruct == CSMAX)
		exit_lowcsmax();
	if (flptr->NumTerms < 3 || strcmp(flptr->Terms[flptr->NumTerms], "THEN"))
	{
		clmes_syntax("Expecting IF <expression> [<operator> <expression>] THEN.");
		return;
	}

	cstructs[curcstruct].ID = makeid("__IF","");
	cstructs[curcstruct].Special = 0;
	cstructs[curcstruct].Type = IfThenElse;

	solvetest(2, flptr->NumTerms - 1, *flptr, 0, file);
}
void anal_else(struct FileLine *flptr, FILE *file)
{
	if (curcstruct == -1 || cstructs[curcstruct].Type != IfThenElse)
	{
		clmes_syntax("ELSE without IF");
		clmes_syntax("(or inside an other control structure).");
		return;
	}
	if (flptr->NumTerms != 1)
	{
		clmes_syntax("Expecting ELSE (without arguments).");
		return;
	}
	cstructs[curcstruct].Special = 1;
	fprintf(file,
		"    JP %sEND__\n"
		"%sTEST__:\n",
		cstructs[curcstruct].ID,
		cstructs[curcstruct].ID);
}
void anal_endif(struct FileLine *flptr, FILE *file)
{
	if (curcstruct == -1 || cstructs[curcstruct].Type != IfThenElse)
	{
		clmes_syntax("ENDIF without IF");
		clmes_syntax("(or inside an other control structure).");
		return;
	}
	if (flptr->NumTerms != 1)
	{
		clmes_syntax("Expecting ENDIF (without arguments).");
		return;
	}
	fprintf(file,
		"%s%s__:\n",
		cstructs[curcstruct].ID,
		cstructs[curcstruct].Special ? "END" : "TEST");
	curcstruct--;
}
void anal_loop(struct FileLine *flptr, FILE *file)
{
	if (++curcstruct == CSMAX)
		exit_lowcsmax();
	
	cstructs[curcstruct].Type = Loop;

	fprintf(file,
		"%sBEGIN__:\n",
		(cstructs[curcstruct].ID = makeid("__LOOP","")));

	if (flptr->NumTerms == 1)
		return;

	if (flptr->NumTerms == 2)
	{
		clmes_syntax("Expecting:");
		clmes_syntax("LOOP [[WHILE|UNTIL] <expression> [<operator> <expression>]].");
		return;
	}
	if (strcmp(flptr->Terms[2], "WHILE") && strcmp(flptr->Terms[2], "UNTIL"))
	{
		clmes_syntax("Expecting:");
		clmes_syntax("LOOP [[WHILE|UNTIL] <expression> [<operator> <expression>]].");
		return;
	}
	
	if (!strcmp(flptr->Terms[2], "WHILE"))
	{
		if (solvetest(3, flptr->NumTerms, *flptr, 0, file))
			return;
	}
	else
		if (solvetest(3, flptr->NumTerms, *flptr, 1, file))
			return;
}

void anal_endloop(struct FileLine *flptr, FILE *file)
{
	if (curcstruct == -1 || cstructs[curcstruct].Type != Loop)
	{
		clmes_syntax("ENDLOOP without LOOP");
		clmes_syntax("(or inside an other control structure).");
		return;
	}
	if (flptr->NumTerms != 1)
	{
		clmes_syntax("Expecting ENDLOOP (without arguments).");
		return;
	}
	fprintf(file,
		"    JP %sBEGIN__\n"
		"%sTEST__:\n",
		cstructs[curcstruct].ID, cstructs[curcstruct].ID);
	curcstruct--;
}

int solvetest(int Start, int End, struct FileLine fl, int jpiftrue, FILE *file)
{
	int i, x;
	int prevop = 0;
	int opatterm[LINE_LENGTH];
	int curopterm = 0;
	char *tempptr;
	char *jptrue;
	char *jpfalse;

	opatterm[0] = Start - 1;
	
	if (ifoperator(fl.Terms[Start]) || ifoperator(fl.Terms[End]))
	{
		clmes_syntax("First or Last keyword of a logical");
		clmes_syntax("test can't be an logical operator.");
		return(-1);
	}
	
	for (i=Start+1; i < End; i++)
		if(x = ifoperator(fl.Terms[i]))
			if (x > 0 && prevop > 0)
			{
				clmes_syntax("Error in logical test:");
				clmes_syntax("Expecting AND or OR.");
				return(-2);
			}
			else if (opatterm[curopterm] == i-1)
			{
				clmes_syntax("Error in logical test:");
				clmes_syntax("Expecting expression between operators.");
				return(-3);
			}
			else
			{
				opatterm[++curopterm] = i;
				prevop = x;
			}

	if (!curopterm)
		if (solve(Start, End, fl, file))
			return(-4);
		else
			fprintf(file,
				"    LD A, H\n"
				"    OR L\n"
				"    JP %s, %sTEST__\n",
				jpiftrue ? "NZ" : "Z",
				cstructs[curcstruct].ID);
	else
	{
		for (i = 0; ++i <= curopterm;)
		{
			if (solve(opatterm[i-1] + 1, opatterm[i] - 1 , fl, file))
				return(-5);
			
			if ((x = ifoperator(fl.Terms[opatterm[i]])) > 0)
			{
				fputs("    PUSH HL\n", file);
				if (solve(opatterm[i] + 1, i == curopterm ? End : opatterm[i+1] - 1 , fl, file))
					return(-6);
				fputs("    POP DE\n", file);
			}			

			if (x == 3 || x == 4)
				fputs("    EX DE, HL\n", file);
			else if (x > 4)
				x -=2;

			if (jpiftrue)
			{
				jptrue = "TEST__";
				jpfalse = "TEMP__";
			}
			else
			{
				jptrue = "TEMP__";
				jpfalse = "TEST__";
			}

			tempptr = NULL;

			if (i != curopterm)
			{
				tempptr = makeid("","NEXT__");
				if (opatterm[i+1] == -1)
					jptrue = tempptr;
				else
					jpfalse = tempptr;
			}
	
			switch (x)
			{
				case -1:
					fprintf(file,
						"    LD A, H\n"
						"    OR L\n"
						"    JP Z, %s%s\n",
						cstructs[curcstruct].ID,
						jpfalse);
					if (i == curopterm)
						if (solve(opatterm[i-1] + 1, opatterm[i] - 1 , fl, file))
							return(-7);
						else
						{
							fputs(	"    LD A, H\n"
								"    OR L\n",
								file);
							fprintf(file,
								"    JP %s, %sTEST__\n",
								jpiftrue ? "NZ" : "Z",
								cstructs[curcstruct].ID);
						}
					break;	
				case -2:
					fprintf(file,
						"    LD A, H\n"
						"    OR L\n"
						"    JP NZ, %s%s\n",
						cstructs[curcstruct].ID,
						jptrue);
					if (i == curopterm)
						if (solve(opatterm[i-1] + 1, opatterm[i] - 1 , fl, file))
							return(-7);
						else
						{
							fputs(	"    LD A, H\n"
								"    OR L\n",
								file);
							fprintf(file,
								"    JP %s, %sTEST__\n",
								jpiftrue ? "NZ" : "Z",
								cstructs[curcstruct].ID);
						}
					break;
				case 1:
					fprintf(file,
						"    LD A, H\n"
						"    CP D\n"
						"    JP NZ, %s%s\n"
						"    LD A, L\n"
						"    CP E\n"
						"    JP NZ, %s%s\n",
						cstructs[curcstruct].ID,
						jpfalse,
						cstructs[curcstruct].ID,
						jpfalse);
					break;
				case 2:
					fprintf(file,
						"    LD A, H\n"
						"    CP D\n"
						"    JP NZ, %s%s\n"
						"    LD A, L\n"
						"    CP E\n"
						"    JP Z, %s%s\n",
						cstructs[curcstruct].ID,
						jptrue,
						cstructs[curcstruct].ID,
						jpfalse);
					break;
				case 3:
					fprintf(file,
						"    LD A, H\n"
						"    CP D\n"
						"    JP C, %s%s\n"
						"    CP D\n"
						"    JP NZ, %s%s\n"
						"    LD A, L\n"
						"    CP E\n"
						"    JP NC, %s%s\n",
						cstructs[curcstruct].ID,
						jptrue,
						cstructs[curcstruct].ID,
						jpfalse,
						cstructs[curcstruct].ID,
						jpfalse);
					break;
				case 4:
					fprintf(file,
						"    LD A, H\n"
						"    CP D\n"
						"    JP C, %s%s\n"
						"    CP D\n"
						"    JP NZ, %s%s\n"
						"    LD A, L\n"
						"    CP E\n"
						"    JP C, %s%s\n",
						cstructs[curcstruct].ID,
						jpfalse,
						cstructs[curcstruct].ID,
						jptrue,
						cstructs[curcstruct].ID,
						jpfalse);
					break;
			}
			if (tempptr)
				fprintf(file,"%s%s:\n", cstructs[curcstruct].ID, tempptr);
			if (i != curopterm && x > 0)
				i++;
		}
		if (x > 0 && jpiftrue)
			fprintf(file,
				"    JP %s%s\n",
				cstructs[curcstruct].ID,
				jptrue);
		fprintf(file, "%sTEMP__:\n", cstructs[curcstruct].ID);
	}
	return(0);
}

int ifoperator(char *str)
{
	if (!strcmp(str, "AND"))
		return(-1);
	if (!strcmp(str, "&&"))
		return(-1);
	if (!strcmp(str, "OR"))
		return(-2);
	if (!strcmp(str, "||"))
		return(-2);
	if (!strcmp(str, "=="))
		return(1);
	if (!strcmp(str, "!="))
		return(2);
	if (!strcmp(str, "<"))
		return(3);
	if (!strcmp(str, ">="))
		return(4);
	if (!strcmp(str, ">"))
		return(5);
	if (!strcmp(str, "<="))
		return(6);
	return(0);
}
