/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bzc.h"

void header(FILE *file)
{
	struct Includes *incptr = incindx;

	if (tasm) fputs(".NOLIST\n\n", file);
	
	for (; incptr; incptr = incptr->Next)
		fprintf(file, "#INCLUDE \"%s\"\n", incptr->File);

	if (tasm) 
	{
		if (!os) fputs("\n#DEFINE  ProgStart    $9D95\n\n", file);
		fputs(".LIST\n", file);
		if (os)
			fputs(".ORG	$0000\n", file);
		else
			fputs(".ORG    ProgStart - 2\n", file);
	}
	else
		if (os)
			fputs(".ORG    $0000\n", file);
		else
			fputs(".ORG    $9D93\n", file);

	if (!os)
	{	fputs(".DB     $0BB, $6D\n\n", file);
		fputs("Start:\n", file);
		fputs("    LD (__RETPointer), SP\n", file);
	}
}

int footer(FILE *file)
{
	int x;

	if (!os) fputs("    RET\n", file);

	// We need first to write functions an variables before ending the file.
	
	x = writefuncs(file);
	if (x)
		return(x);

	writevars(file);
	
	if (tasm)
		fputs("\n.END\n.END\n", file);
	
	if (file != stdout && file != stderr) fclose(file);
	return(0);
}

int writefuncs(FILE *file)
{
	int x;
	struct SaveTab *temptab;
	struct FileLine *flptr;

	for (temptab = tabptr; temptab; temptab = temptab->Next)
	
		if (temptab->Type == function_user && temptab->Used > 0)
		{
			curfunc = temptab->Name;	

			fprintf(file, "\n%s:\n", temptab->Name);
			
			flptr = gotoline(temptab->Line);
			
			if (temptab->Args)
			{
				if (temptab->Args + 2 != flptr->NumTerms)
				{
					clmes_syntax("Number of declared arguments differs from");
					clmes_syntax("number of Used Arguments.");
					return(flptr->Line);
				}
								
				fputs("    POP BC\n", file);
				
				for (x = flptr->NumTerms; x > 2; x--)
				
					switch (popvar(flptr->Terms[x], file))
					{
						case 0:
							break;
						case 1:
							totable(var, flptr->Terms[x],
								"", 0, temptab->Line, 4);
							if (!popvar(flptr->Terms[x], file))
								break;
						default:
							clmes_notdecl(flptr->Terms[x]);
							return(line);
					}

				fputs("    PUSH BC\n", file);
			}

			x = compilefunction(temptab, file);

			if (x)
				return(x);
			
			fputs("    RET\n", file);
			
			temptab->Used = -1;

			// Check dependencies:

			if(x = writefuncs(file)) return(x);
			break;
		}

	return(0);
}

void writevars(FILE *file)
{
	int i;
	int compile;
	int *arr;
	char *str;
	struct SaveTab *temptab;

	// We're going to write every 'var' entry to the file:

	for (temptab = tabptr; temptab; temptab = temptab->Next)
	{
		// Check if it should be compiled in (is it a VAR, which type, is it used?)
		
		if (temptab->Used > 0)
			compile = 1;
		else if (temptab->Used < 0)
			compile = 0;
		else if (temptab->Next && temptab->Next->Lastline == 5 && !strcmp(temptab->Next->Realname, temptab->Name))
			compile = 1;

		if (compile && temptab->Type == var && temptab->Lastline - 1 && temptab->Lastline != 5)
		{
			fprintf(file, "\n%s:\n", *temptab->Realname ? temptab->Realname \
							: temptab->Name);
			
			if (temptab->Lastline != 2)	// it's NOT a string.
			
				fprintf(file, ".D%c %d\n", temptab->Lastline == 3 ? 'B' : \
					'W', temptab->Args);
			else
			{
				if (*(str = gotoline(temptab->Line)->Terms[temptab->Args]))
					for (; *str; fprintf(file, ".DB %d\n", *str++));

				fprintf(file, ".DB 0\n");
			}
		temptab->Used = -1;
		}
	}

	// We're going to write every 'array' entry to the file:
	
	for (temptab = tabptr; temptab; temptab = temptab->Next)

		if (temptab->Type == array && temptab->Used > 0)
		{
			fprintf(file, "\n%s:\n", temptab->Name);

			if (!temptab->Lastline)
				for (i = 0; i < temptab->Args; i++)
					fputs(".DW 0\n", file);
			else
			{
				arr = ptrs[temptab->Lastline];
				for (i = 0; i < temptab->Args; i++)
					fprintf(file, ".DW %d\n", *arr++);
			}
		temptab->Used = -1;
		}
}

struct FileLine *load(FILE *infile, struct FileLine *parent)
{
	// This is the second load(), rewritten for performence and a
	// more bugless source.
	
	int filename;
	int backupline;
	int chars = 0;
	char c, testc;
	char *bakname;
	char *curchar;
	char readline[LINE_LENGTH];
	char include_file[FNAME_LENGTH];
	FILE *incfile;
	struct FileLine *flptr, *oldline = NULL;

	
	// First, check if we are the main source file.
	
	if (strlen(progname) > FNAME_LENGTH)
		exit_lowfname();

	if (!parent)
		flindex = flptr = (struct FileLine *) getmem(sizeof(struct FileLine));
	else
		flptr = parent;

	if (!flptr)
		exit_nomem();

	// Find a new filename entry:
	
	for (filename = 0; filename < MAX_FILES && Files[filename]; filename++);
	
	if (filename == MAX_FILES)
		exit_maxfiles();
	
	if (!(Files[filename] = (char *) getmem(FNAME_LENGTH)))
		exit_nomem();

	flptr->FileName = filename;

	strcpy(Files[filename], progname);

	flptr->Line = 1;
	flptr->NumTerms = 0;
	
	curchar = readline;

	// Start reading the file
	
	for(;;)
	{
		switch (c = getc(infile))
		{
			case '<':
			case '>':
			case '!':
			case '+':
			case '-':
			case '*':
			case '/':
			case '%':
			case '=':
			case '(':
			case ')':
				if (c != '/' || chars || !flptr->NumTerms || strcmp(flptr->Terms[flptr->NumTerms], "/"))
				{
					if ((!chars && c == '=' && flptr->NumTerms) && (
						!strcmp(flptr->Terms[flptr->NumTerms], "<") ||
						!strcmp(flptr->Terms[flptr->NumTerms], ">") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "!") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "=") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "+") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "-") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "*") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "/") ||
						!strcmp(flptr->Terms[flptr->NumTerms], "%")))
					{
						curchar = readline;
						*curchar++ = *flptr->Terms[flptr->NumTerms];
						chars++;
					}
					else 
					{
						if (chars) // Close current term.
						{
							*curchar = '\0';
					
							if (chars > LINE_LENGTH)
								exit_lowline();
							
							if(!(flptr->Terms[flptr->NumTerms] = (char *) getmem(chars + 1)))
								exit_nomem();
					
							flptr->Quoted[flptr->NumTerms] = 0;

							strcpy(flptr->Terms[flptr->NumTerms], readline);
							chars = 0;
						}
						curchar = readline;
						flptr->NumTerms++;
					}
					*curchar++ = c;
					*curchar = '\0';
					chars++;

					if (chars > LINE_LENGTH)
						exit_lowline();

					if(!(flptr->Terms[flptr->NumTerms] = (char *) getmem(chars + 1)))
						exit_nomem();
					
					flptr->Quoted[flptr->NumTerms] = 0;

					strcpy(flptr->Terms[flptr->NumTerms], readline);
					chars = 0;
					break;
				}

				flptr->NumTerms--;

			case ';':
				if (!chars)
				{
					*curchar = '\0';
					for (; c != '\n' && c != EOF; c = getc(infile));
				}
				else
				{
					*curchar++ = c;
					chars++;
					break;
				}
			case '	':	// TAB
			case ' ':
			case EOF:
			case '\n':
				
				if (chars)	// Close current term.
				{
					*curchar = '\0';
					
					if (chars > LINE_LENGTH)
						exit_lowline();

					if(!(flptr->Terms[flptr->NumTerms] = (char *) getmem(chars + 1)))
						exit_nomem();
					
					flptr->Quoted[flptr->NumTerms] = 0;

					strcpy(flptr->Terms[flptr->NumTerms], readline);
					chars = 0;
				}
				
				if (c != '\n') 
					break;
				
				if (flptr->NumTerms && !strcmp(flptr->Terms[1], "#include"))
				{
					if (flptr->NumTerms != 2)
						if (flptr->NumTerms != 4 ||
							strcmp(flptr->Terms[2], "<") ||
							strcmp(flptr->Terms[4], ">"))

							exit_inc();
										
					bakname = progname;
					progname = flptr->Terms[flptr->NumTerms == 2 ? 2 : 3];
					
					if (!(incfile = fopen(progname, "r")))
					{
						if ((strlen(HEADER_PATH) + strlen(progname)) > FNAME_LENGTH)
							exit_lowfname();

						strcpy(include_file, HEADER_PATH);
						strcat(include_file, progname);
						if (!(incfile = fopen(include_file, "r")))
						{
							clmes_old("Error while opening include file.");
							exit_io();
						}
					}

					backupline = line;
					
					flptr = load(incfile, flptr);
					
					fclose(incfile);
					
					progname = bakname;
					line = backupline;
					
					if (!flptr)
						return(NULL);
				} 
				else if (flptr->NumTerms > 1 && !strcmp(flptr->Terms[1], "INCLUDE"))

					if (flptr->NumTerms == 2)	
					{
						addinc(flptr->Terms[2]);
						flptr->NumTerms = 0;
					}
					else
					{
						clmes_syntax("Expecting INCLUDE \"include_file\",");
						
						if (flptr->NumTerms == 1)
							clmes_syntax("but RETURN at \"include_file\".");
						else
							clmes_syntax("but no RETURN after \"include_file\".");
						
						return(NULL);
					}
												
				if (flptr->NumTerms)
				{
					// Allocating space for a new line:
	
					if(!(flptr->Next = (struct FileLine *) getmem(sizeof(struct FileLine))))
						exit_nomem();
					
					oldline = flptr;
					flptr = flptr->Next;
					flptr->FileName = filename;
				}
				
				flptr->NumTerms = 0;
				flptr->Line = line++;
				break;
				
			case '\"':
			/*	if (c == '<')
					testc = '>';
				else	*/
					testc = '\"';					
				
				if (!chars)
				{
					curchar = readline;
					flptr->NumTerms++;
				}
				
				for (c = fgetc(infile); c != testc && c != EOF && c != '\n'; c = fgetc(infile))
				{
					*curchar++ = c;
					chars++;
				}				
												
				if (c != testc)
				{
					clmes_syntax("End of line before end-of-keyword character (\" or >).");
					return(NULL);
				}
				
				// Close the current term (same as with ' ')
				
				*curchar = '\0';
					
				if (chars > LINE_LENGTH)
					exit_lowline();

				if(!(flptr->Terms[flptr->NumTerms] = (char *) getmem(chars + 1)))
					exit_nomem();

				strcpy(flptr->Terms[flptr->NumTerms], readline);
				flptr->Quoted[flptr->NumTerms] = 1;
				chars = 0;
				
				break;
				
			case '\r':		// Some DOS / WINDOWS files have these anoying chars
				break;		// at the end of every line.
		
			default:
				if (!chars)
				{
					curchar = readline;
					flptr->NumTerms++;
				}
				*curchar++ = c;
				chars++;
				break;
		}
		
		if (c == EOF)
		{
			if (!flptr->NumTerms)
				if (!oldline)
				{
					clmes_old("Source or include file is empty.");
					return(NULL);
				}
				else
					flptr = oldline;
				
			flptr->Next = NULL;
			break;
		}
	}			

	if (verbose) clmes_old("Succesfully loaded !!");
	
	return(flptr);
}
