/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>    // for strcmp()

#include "bzc.h"

int examline(FILE *file)
{	 
	char str[LINE_LENGTH];
	struct FileLine *flptr = flindex;
	
	for (line = 0; flptr; flptr = flptr->Next)
	{
		line++;
		wholeline(str, flptr);
		
#ifdef CMD_DEBUG
		clmes_debug_ds("Reading line %3d: %s", flptr->Line, str, 0);
#ifdef MEMLINE_DEBUG
		clmes_debug_ds("Next pointer is: %p", flptr->Line, flptr->Next, 1);
#endif
#endif
		if (comflag == Code)
			if (!strcmp(flptr->Terms[1], "FEND"))
				if (flptr->NumTerms == 1)
					flastline();
				else
				{
					clmes_syntax("Expecting RETURN after FEND.");
					return(line);
				}
			else	;	// We're just seeking...
			
		else if (AsmCom)
			if (!strcmp(flptr->Terms[1], "ENDASM"))
				if (flptr->NumTerms == 1)
					AsmCom = 0;
				else
				{
					clmes_syntax("Expecting RETURN after ENDASM.");
					return(line);
				}
			else
				fprintf(file, "%s\n", mkasm(str));
		else
			// Compile this line.
		{
			cmperr = 0;
			examterm(flptr, file);

			if (cmperr)
				return(line);
		}
	}
	return(0);
}

int equterm(struct FileLine *flptr, char *equ, int term_nr)
{
	char temp[VAR_LENGTH];

	term(flptr, temp, term_nr);
	
	if (!*temp)
		return (-1);
	if (!strcmp(equ, temp))
		return (0);
	return(1);
}

void examterm(struct FileLine *flptr, FILE *file)
{
	char *place = flptr->Terms[1]; 

#ifdef ANAL_DEBUG
	clmes_debug_s("Analyzing keyword: %s.", place, 0);
#endif

        if (!strcmp(place, "REG8"))
		newreg8(flptr);
	else if (!strcmp(place, "REG16"))
        	newreg16(flptr);
        else if (!strcmp(place, "FUNCTION"))
		if (comflag == Noflag)
			newfunc(flptr);
		else
			clmes_syntax("FUNCTION keyword inside a program routine.");
	else if (!strcmp(place, "ROMCALL"))
        	newrcall(flptr, file);
	else if (!strcmp(place, "MEM"))
		newarray(flptr);
	else if (!strcmp(place, "SPRITE"))
		newsprite(flptr);
	else if (!strcmp(place, "SYSVAR") || \
		 !strcmp(place, "STRING") || \
		 !strcmp(place, "INT") || \
		 !strcmp(place, "BYTE") || \
		 !strcmp(place, "CONST"))
		newvar(flptr, file);
	else if (!strcmp(place, "FORCE"))
		if (comflag != Noflag)
			forcevar(flptr);
		else
			clmes_syntax("FORCE keyword outside a program routine.");
	else if (!strcmp(place, "PRINTF"))
		if (comflag != Noflag)
			anal_printf(flptr, file);
		else
			clmes_syntax("PRINTF keyword outside a program routine.");
	else if (!strcmp(place, "RETURN"))
		if (comflag != Noflag)
			anal_return(flptr, file);
		else
			clmes_syntax("RETURN keyword outside a program routine.");
	else if (!strcmp(place, "EXIT"))
		if (comflag != Noflag)
			anal_exit(flptr, file);
		else
			clmes_syntax("EXIT keyword outside a program routine.");
	else if (!strcmp(place, "REPEAT"))
		if (comflag != Noflag)
			anal_repeat(flptr, file);
		else
			clmes_syntax("REPEAT keyword outside a program routine.");
	else if (!strcmp(place, "ENDREPEAT"))
		if (comflag != Noflag)
			anal_endrepeat(flptr, file);
		else
			clmes_syntax("ENDREPEAT keyword outside a program routine.");
	else if (!strcmp(place, "FOR"))
		if (comflag != Noflag)
			anal_for(flptr, file);
		else
			clmes_syntax("FOR keyword outside a program routine.");
	else if (!strcmp(place, "NEXT"))
		if (comflag != Noflag)
			anal_next(flptr, file);
		else
			clmes_syntax("NEXT keyword outside a program routine.");
	else if (!strcmp(place, "LOOP"))
		if (comflag != Noflag)
			anal_loop(flptr, file);
		else
			clmes_syntax("LOOP keyword outside a program routine.");
	else if (!strcmp(place, "ENDLOOP"))
		if (comflag != Noflag)
			anal_endloop(flptr, file);
		else
			clmes_syntax("ENDLOOP keyword outside a program routine.");
	else if (!strcmp(place, "IF"))
		if (comflag != Noflag)
			anal_if(flptr, file);
		else
			clmes_syntax("IF keyword outside a program routine.");
	else if (!strcmp(place, "ELSE"))
		if (comflag != Noflag)
			anal_else(flptr, file);
		else
			clmes_syntax("ELSE keyword outside a program routine.");
	else if (!strcmp(place, "ENDIF"))
		if (comflag != Noflag)
			anal_endif(flptr, file);
		else
			clmes_syntax("ENDIF keyword outside a program routine.");
	else if (!strcmp(place, "ASM"))
		if (comflag != Noflag)
			anal_asm(flptr, file);
		else
			clmes_syntax("ASM keyword outside a program routine.");
	else if (!strcmp(place, "ENDASM"))
		
		clmes_syntax("ENDASM outside an ASM block.");

	else if (!strcmp(place, "FSTART"))
	
		if (comflag == Noflag)
			anal_code(flptr, file);
		else
			clmes_syntax("FSTART keyword inside a program routine.");
	
	else if (!strcmp(place, "FEND"))

		if (comflag == Code)
			anal_endcode(flptr, file);
		else
			clmes_syntax("FEND outside a function.");

	else if (!strcmp(place, "START"))
	
		if (comflag == Noflag)
			anal_start(flptr, file);
		else
			clmes_syntax("START keyword inside a program routine.");

	else if (!strcmp(place, "END"))
			
		if (comflag == Start)
			anal_endcode(flptr, file);
		else
			clmes_syntax("END outside an START routine.");
	else
		if (anal_varfunc(flptr, file) == 1)
		{
			clmes_syntax("~VAR or FUNCTION \"");
			fprintf(stderr, "%s\" not declared.\n", place);
		}
		else if (comflag == Noflag)
			clmes_syntax("Accessing an VAR or FUNCTION outside a program routine.");

#ifdef ANAL_DEBUG
	if (!cmperr)
		clmes_debug("Line accepted and compiled.", 1);
	else
		clmes_debug_ds("Syntax Error %d!", cmperr, "", 1);
#endif
}

int todigit(char *str)
{
	char highest = '9';
	int times = 10;
	
	int digit = 0;

	switch (*str)
	{
		case '$':
			for (str++; *str && *str != '\n'; str++)
			{
				digit *= 16;
	
				if (*str >= '0' && *str <= '9')
					digit += (*str - 48);
				else if (*str >= 'A' && *str <= 'F')
					digit += (*str - 55);
				else if (*str >= 'a' && *str <= 'f')
					digit += (*str - 87);
				else
					return(-1);
			}
			break;
		case '#':
			highest = '1';
			times = 2;
			str++;
		default:
			for (; *str && *str != '\n'; str++)
				if (*str >= '0' && *str <= highest)
				{
					digit *= times;
					digit += (*str - 48);
				}
				else
					return(-1);
			break;
	}
	return(digit);
			
}

int checkex(char *name, struct FileLine *flptr)
{
	int i = 0;
	struct SaveTab *extab;
	
	if (extab = findtable(name, 1))
	{
		clmes_syntax_dec(name, extab->Line);
		return(1);
	}
	
	for (; i < RESERVED_NUM; i++)
		if (!strcmp(name, KeyWords[i]))
		{
			clmes_syntax_keyw(name);
			return(2);
		}
	
	return(0);
}

char *mkasm(char *str)
{
	char *asmstr;
	char *c, *ac;
	char *temp;
	char x;
	
	ac = (asmstr = getmem(LINE_LENGTH));
	
	for (c = str; *c; c++);
	if (*--c != ':' && *str != '.')
	{
		*ac++ = ' ';
		*ac++ = ' ';
		*ac++ = ' ';
		*ac++ = ' ';
	}
	
	for (c = str; *ac++ = *c++;)
		if (*c == '(')
		{
			temp = c;
			for (c++; *c == ' ' || *c == '	'; c++);
			if (*c++ == 'I')
				if ((x = *c) == 'X' || *c == 'Y')
				{
					for (c++; *c == ' ' || *c == '	'; c++);
					if (*c == ')')
					{
						*ac++ = '(';
						*ac++ = 'I';
						*ac++ = x;
						*ac++ = '+';
						*ac++ = '0';
					}
					else
						c = temp;
				}
				else
					c = temp;
			else
				c = temp;
		}
		else if (*c == ' ' || *c == '	')
		{
			if (*(c-1) == '(' || *(c+1) == ')' ||
				( *(c-1) == ')' && *(c+1) == ','))
				c++;
		}
		
	return(asmstr);
}

int toHLpp(struct FileLine fl, int Start, int Type, FILE *file, int SaveDE)
{
	int i, j = 0;
	int x, y;
	struct SaveTab *stab;

	switch (Type)
	{
		case -1:
			fprintf(file, "    LD HL, %s\n", fl.Terms[Start]);
			break;
		case -2:
			if (SaveDE) fputs("    PUSH DE\n", file);
			stab = findtable(fl.Terms[Start], 0);
			
			for (i = Start; i < stab->Args + Start;)
				if (strcmp(fl.Terms[++i + j], "("))
					pushvar(&fl, file, i + j);
				else
				{
					for (x = y = 1; x > 0; y++)
						if (!strcmp(fl.Terms[i + j + y], "("))
							x++;
						else if (!strcmp(fl.Terms[i + j + y], ")"))
							x--;

					x = solve(i + j + 1, i + j + --y - 1, fl, file);
					if (x) return(x);
					fputs("    PUSH HL\n", file);
					j += y;
				}		
			
			fprintf(file, "    CALL %s\n", stab->Name);
			if (SaveDE) fputs("    POP DE\n", file);
			break;
		case -3:
			stab = findtable(fl.Terms[Start], 0);
			toHL(stab, file, 0);
			break;			
		case -4:
			globalvar = 1;
			stab = totable(var, makeid("__STRING",""), "", Start, line, 2);
			fprintf(file, "    LD HL, %s\n", stab->Name);
			stab->Used++;
			break;
		default:
			if (SaveDE) fputs("    PUSH DE\n", file);
			i = solve(Start + 1, Type - 1, fl, file);
			if (SaveDE) fputs("    POP DE\n", file);
			return(i);
	}
	

	return(0);
}

