/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>	// For exit()
#include <ctype.h>      // For toupper()
#include <string.h>	// For strcmp() and strcpy()

#define DECL_VARS	// Declare vars in bzc.h (may happen only once)

#include "bzc.h"

char realprogname[] = PROGNAME;
int prevline = -99;

int compiler(int argc, char *argv[]) 
{
	
	int result;
	FILE *infile, *outfile;
	char uppercase[CL_ARG_LENGTH];
	
	progname = realprogname;
	if (verbose) fprintf(stderr, "\n");

	switch (argc) 
	{
		case 3:
			break;
#ifdef OUTPUT_DEVICE
#ifdef INPUT_DEVICE
		case 1:
			argv[1] = INPUT_DEVICE;
#endif		
		case 2:
			argv[2] = OUTPUT_DEVICE;
			break;
#endif
		default:
			clmes_old("Wrong Syntax!");
			clmes_old("~Usage: ");
			fprintf(stderr, "%s <input file> <output file>\n\n", \
								argv[0]);
			return(0);
	}
		
	if (strlen(argv[1]) > CL_ARG_LENGTH)
		exit_lowcl();

	transtoup(uppercase, argv[1]);
	
	if (!strcmp(uppercase, "STANDART INPUT"))
		infile = stdin;
	else
		if ((infile = fopen(argv[1],"r")) == NULL)
		{
			clmes_old("Unable to open input file!");
			exit_io();
		}
		
	if (strlen(argv[2]) > CL_ARG_LENGTH)
		exit_lowcl();
	transtoup(uppercase, argv[2]);
	
	if (!strcmp(uppercase, "STANDART OUTPUT"))
		outfile = stdout;
	else if (!strcmp(uppercase, "STANDART ERROR OUTPUT"))
		outfile = stderr;
	else
		if ((outfile = fopen(argv[2],"w")) == NULL)
		{
			clmes_old("Unable to create output file!");
			clmes_old("Check if the output directory exist and");
			clmes_old("if there is enough space left.\n");
			exit_io();
		}

	if (verbose) 
	{
		clmes_old("~Compiling ");
		fprintf(stderr, "%s to %s\n", argv[1], argv[2]);
	}
	
	progname = argv[1];
	initmem();
	
	// Calling the actual compiling function:
	
	result = compile(infile, outfile);
	
	if (infile != stdin) fclose(infile);
	progname = realprogname;
	
	if (result)
	{
		clmes_old("Exiting with errors, see line numbers above.");
		result = SynErr;

#ifdef DUMP_MEM_ON_ERRORS

		dumpmem();
#endif
	}
	else	
		if (verbose) clmes_old("Compile succesfull !!");
	
	if (verbose) fprintf(stderr, "\n");
	
	freemem();
	return(result);
}


void clmes(char *strngf, enum MesType typemes, int lnspc, int digit1, int digit2, void *pointer1)
{
	int i;
	int atline;
	char str[LINE_LENGTH];
	char *strng;
	char *fname, *atfile;
	struct FileLine *temp;

	strng = (*strngf == '~' ? strngf + 1 : strngf);
		
	if (lnspc)
		if(temp = gotoline(lnspc))
		{
			lnspc = temp->Line;
			fname = Files[temp->FileName];
		}
		else
		{
			fname = "System line";
		}		
	
	if (typemes > 0 && typemes < 6)
		if (line != prevline) //This only works as long as lnspc == line
		{
			prevline = line;
			if (temp) 
			{
				wholeline(str, temp);
				fprintf(stderr,"%s: %3d: \"%s\"\n", fname, lnspc, str);
			}
		}

	switch (typemes)
	{
		case 0:		// file
			fprintf(stderr, "%s: %s", progname, \
				*strng == '~' ? strng + 1 : strng);
			break;
		case 1:		// fline
			fprintf(stderr, "%s: %3d: %s", fname, lnspc, \
				*strng == '~' ? strng + 1 : strng);
			break;
		case 2:		// flsyn
			fprintf(stderr, "%s: %3d: Error: %s", fname, \
				lnspc, *strng == '~' ? strng + 1 : strng);
			cmperr = -1;
			break;
		case 3:		// flsyn_dec
			if (!digit2)
			{
				atline = 0;
				atfile = "<The Default Compiler Settings>";
			}
			else if (temp = gotoline(digit2))
			{
				atline = temp->Line;
				atfile = Files[temp->FileName];
			}
			else
			{
				atline = digit2;
				atfile = "System Line";
			}
									
			fprintf(stderr, "%s: %3d: Error: \"%s\" already declared at line %d of file %s.", \
				fname, lnspc, strng, atline, atfile);
			cmperr = -1;
			break;
		case 4:
			fprintf(stderr, "%s: %3d: Error: User defined name \"%s\" is a KEYWORD.", \
				fname, lnspc, strng);
			cmperr = -1;
			break;
		case 5:
			fprintf(stderr, "%s: %3d: Error: \"%s\" is not declared (or can't store data).", \
				fname, lnspc, strng);
			cmperr = -1;
			break;
		case 6:		// memdebug_dp
			fprintf(stderr, "MEMORY DEBUG: ");
			fprintf(stderr, strng, digit1, pointer1);
			break;
		case 7:		// memdebug_cdpd
			fprintf(stderr, "MEMORY DEBUG: ");
			fprintf(stderr, strng, digit1, digit1, pointer1, digit2);
			break;
		case 8:
		case 9:		// debug*
		case 10:
			if (lnspc)
			{
				for (i = 0; i < lnspc; i++)
					fprintf(stderr, "  ");
				fprintf(stderr, "     >>> ");
			}
			else
				fprintf(stderr, "DEBUG: ");
				
			if (typemes == 8)	// debug
				fprintf(stderr, strng);
			else if (typemes == 9)	// debug_s
				fprintf(stderr, strng, pointer1);
			else			// debug_ds
				fprintf(stderr, strng, digit1, pointer1);
			break;
	}

	if (*strngf && *strngf != '~')
		fprintf(stderr, "\n");
}


int compile(FILE *infile, FILE *outfile)
{
	int result;	
		
	if (!(load(infile, NULL)))
		return(1);
			
	if (verbose) fprintf(stderr, "\n");
	
	header(outfile);
	
	if (result = (examline(outfile) || footer(outfile)))
	{
		fprintf(stderr, "\n");
		clmes_old("Compiling (Syntax) error in source or include files.");
		result = 6;
	}
	
	return(result);
}

void transtoup(char *output, char *str)
{
        char orig[VAR_LENGTH], *ptr = orig;

	for (strcpy(orig, str); *ptr; ptr++)
		*ptr = toupper(*ptr);
	
        strcpy(output, orig);
}

void exit_fail(enum ExitCodes cexit)
{
	char *message;
	char *message2 = "";
	
	switch (cexit)
	{
		case 1:		// NoMem
			message = "Could not allocate enough memory (No free memory left).";
			break;
		case 2:		// ioErr
			message = "General IO Error (File read / write error).";
			break;
		case 3:		// lowMaxBlocks
			message = "MAXBLOCKS (defined in bzc.h) is too low";
			message2 = "(Make your program smaller).";
			break;
		case 4:		// LowMaxFiles
			message = "MAX_FILES (defined in bzc.h) is too low";
			message2 = "(Use less include files).";
			break;
		case 5:		// LowMaxPointers
			message = "MAXPTRS (defined in bzc.h) is too low";
			message2 = "(Use les ARRAY's or SPRITE's).";
			break;		
		case 6:		// WrongInc
			message = "#include with too many (or without any) arguments.";
			break;
		case 7:		// NoMath
			message = "Mathematical functions not loaded";
			message2 = "(use math.h or asmlib.h as header).";
			break;
		case 8:		// LowCL
			message = "CL_ARG_LENGTH (defined in bzc.h) is too low";
			message2 = "(Your source or destination file string is too long).";
			break;		
		case 9:		// LowSpriteLines
			message = "MAX_SPRITE_LINES (defined in bzc.h) is too low";
			message2 = "(One of your SPRITE's is too big).";
			break;
		case 10:	// LowFnameLength
			message = "FNAME_LENGTH (defined in bzc.h) is too low";
			message2 = "(one of your source or include file-name is too long).";
			break;
		case 11:	// LowLine
			message = "LINE_LENGTH (defined in bzc.h) is too low";
			message2 = "(one line in your source is too long).";
			break;
		case 12:	// LowCS
			message = "CSMAX (defined in bzc.h) is too low";
			message2 = "(Have less control-structures at a time.).";
			break;

	}
	
	fprintf(stderr, "%s: Compiler error: %s\n", realprogname, message);

	if (*message2)
		fprintf(stderr, "%s: Compiler error: %s\n", realprogname, message2);

	switch(cexit)
	{
		case 3:
		case 4:
		case 5:
		case 8:
		case 9:
		case 10:
		case 11:
		case 12:
			fprintf(stderr, "%s: Compiler error: %s\n", realprogname,
				"If this happends often, please write an email with this");
			fprintf(stderr, "%s: Compiler error: %s\n", realprogname,
				"problem to: bugs@hofhom.nl (subject \"BZC\").");
	}		

	fprintf(stderr, "Exiting now.\n\n");
#ifdef MEMORY_DUMP_ON_EXIT_FAILURE

	dumpmem();
#endif
	freemem();
	
	exit(cexit);
}

int poweroftwo(int x)
{
	int i, j = 2;
	
	if (!x)
		return(1);
	else if (x == 1)
		return(2);
	
	for (i = 1; i < x; i++)
		j = j * 2;
	
	return(j);
}
