/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bzc.h"

void freemem(void)
{
	int i, endindex;

	endindex = memloc / BLOCKSIZE + 1;
	
	if (memloc % BLOCKSIZE)
		endindex++;
	
	for (i = 0; i < endindex; free(memptr[i++]));

}

void initmem(void)
{
	incindx = NULL;
	line = 2;
	cmperr = 0;
	tabptr = NULL;
	comflag = Noflag;
	memloc = 0;
	curptr = 0;
	idnum = 0;
	curfunc = "";
	globalvar = 0;
	pushes = 0;
	curcstruct = -1;

	// Load the register to the memory.
	
	init_reg();
}

void wholeline(char *str, struct FileLine *flptr)
{
	int i;
	char ln[LINE_LENGTH];

	strcpy(ln, flptr->Terms[1]);
	
	for(i = 1; i < flptr->NumTerms; i++)
	{
		strcat(ln, " ");
		strcat(ln, flptr->Terms[i + 1]);
	}
	
	strcpy(str, ln);
}

void term(struct FileLine *flptr, char *target, int term_nr)
{
	if (flptr->NumTerms < term_nr)
		*target = '\0';
	else
		strcpy(target, flptr->Terms[term_nr]);
}

struct FileLine *gotoline(int line_nr)
{
	int i;
	struct FileLine *flptr = flindex;

	for (i = 1; i++ < line_nr && flptr; flptr = flptr->Next);

	if (i < line_nr) 
		return(NULL);
	
	return(flptr);
}

void addinc(char *fname)
{
	struct Includes *incptr = incindx;

	if (incindx)
	{
		for (; incptr->Next; incptr = incptr->Next);

		incptr->Next = (struct Includes *) getmem(sizeof(struct \
								Includes));

		if (!incptr->Next)
			exit_nomem();
		
		incptr = incptr->Next;
	}
	else if (incindx = (struct Includes * )getmem(sizeof(struct Includes)))
		incptr = incindx;
	else
		exit_nomem();

	incptr->Next = NULL;
	incptr->File = fname;
}

void *getmem(int bytes)
{
	int index, offset;

	index = memloc / BLOCKSIZE;
	offset = memloc % BLOCKSIZE;
	memloc += bytes;
	
	if (!offset || memloc >= BLOCKSIZE)
	{
		if (offset)
			index++;
		
		if (index == MAXBLOCKS)	
			exit_maxblcks();

#ifdef GETMEM_DEBUG	
		fprintf(stderr, "DEBUG: getmem(): Allocating %d bytes at %p.\n", bytes, memptr[index]);
#endif
		return(memptr[index] = malloc(BLOCKSIZE));
	}

#ifdef GETMEM_DEBUG
	
	fprintf(stderr, "DEBUG: getmem(): Allocating %d bytes at %p.\n", bytes, memptr[index] + offset);
#endif	
	return((void * ) (((char *) memptr[index]) + offset));	
}

int ptrtoint(int *ptr)
{
	if (++curptr == MAXPTRS)
		exit_lowarr();
	
	ptrs[curptr] = ptr;
	
	return(curptr);
}


#ifdef MEMORY_DUMP_ON_EXIT_FAILURE

void dumpmem()
{
	unsigned long i;
	FILE *dumpfile;

	dumpfile = fopen("memory.dmp","w");

	if (dumpfile)
		clmes_old("Dumping Memory to file \"memory.dmp\".");
	else
		clmes_old("Could not dump memory, I/O error.");
	
	for (i = 0; i < memloc; i++)
		fputc(*( (char *) memptr[i / BLOCKSIZE] + i % BLOCKSIZE), dumpfile);

	fclose(dumpfile);
}

#endif
