/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>

#include "bzc.h"

void newreg8(struct FileLine *flptr)
{
	char *regname;
	
#ifdef REG_DEBUG

        clmes_debug("NEW REG8 keyword found.", 2);
#endif
	if (flptr->NumTerms > 2)
	{
		clmes_syntax("Expecting RETURN after NEW REG8 <register>.");
		return;
	}
	
	if (flptr->NumTerms < 2)
	{
		clmes_syntax("Expecting NEW REG8 <register>, but RETURN at <register>'s place.");
		return;
	}
	
	regname = flptr->Terms[2];
	
	if (checkex(regname, flptr))
		return;
	
	totable(reg8, regname, "", 0, line, 0);
}

void newreg16(struct FileLine *flptr)
{
	char *regname;
	
#ifdef REG_DEBUG

        clmes_debug("NEW REG16 keyword found.", 2);
#endif
	if (flptr->NumTerms > 2)
	{
		clmes_syntax("Expecting RETURN after NEW REG16 <register>.");
		return;
	}
	
	if (flptr->NumTerms < 2)
	{
		clmes_syntax("Expecting NEW REG16 <register>, but RETURN at <register>.");
		return;
	}
	
	regname = flptr->Terms[2];
	
	if (checkex(regname, flptr))
		return;
	
	totable(reg16, regname, "", 0, line, 0);
}

void newarray(struct FileLine *flptr)
{
	int num;

	if (flptr->NumTerms != 3 || (num = todigit(flptr->Terms[3])) == -1)
	{
		clmes_syntax("Expecting MEM <array_name> <length_in_bytes>.");
		return;
	}
	if (checkex(flptr->Terms[2], flptr))
		return;
	
	totable(array, flptr->Terms[2], "", (num+1) / 2, line, 0);

}

void newfunc(struct FileLine *flptr)
{
	int rmcall;
	int args;
	int checkargs = 1;
	int decline = 0;
	char funcname[VAR_LENGTH];
	char rcname[VAR_LENGTH];

	struct SaveTab *rcalltab;
	
	enum SaveType temptype;
		
#ifdef NEW_FUNC_DEBUG

        clmes_debug("FUNCTION keyword found.", 2);
#endif
	if (equterm(flptr, "", 2) == -1)
		clmes_syntax("Expecting FUNCTION <function_name> *, but RETURN at <function_name>.");
	else
	{
		term(flptr, funcname, 2);
		if (checkex(funcname, flptr))
			return;
		rmcall = equterm(flptr, "ROMCALL", 3);
		
		if (!rmcall)
			if (equterm(flptr, "", 4) == -1)
			{
				clmes_syntax("Expecting FUNCTION <function_name> ROMCALL <romcall_name>,");
				clmes_syntax("but RETURN at <romcall_name>.");
				return;
			}
			else	// FUNCTION <function_name> ROMCALL *
			{
				temptype = function_rcall;
				decline = line;
					if(!(rcalltab = findtable(flptr->Terms[4], 0)))
				{
					clmes_syntax("Expecting FUNCTION <function_name> ROMCALL <romcall_name>,");
					clmes_syntax("but <romcall_name> is not declared with ROMCALL <romcall_name>.");
					return;
				}
				strcpy(rcname, *rcalltab->Realname ? rcalltab->Realname : rcalltab->Name);
			}
		else
		{
			temptype = function_user;
			*rcname = '\0';
			
			
			switch (equterm(flptr, "USER", 3))
			{
				case -1:
					args = 0;
					checkargs = 0;
				case 0:
					break;
				case 1:
					if (strcmp(flptr->Terms[3], "ARG") || \
						flptr->NumTerms != 4 ||
						(args = todigit(flptr->Terms[4])) == -1)
					{
						clmes_syntax("Expecting FUNCTION <function_name> [ ARG <number> ].");
						return;
					}
					checkargs = 0;
			}
		}
		// * [ ARGS * ]
		if (checkargs) switch (equterm(flptr, "ARG", 5 - rmcall))
		{
			case 1:
				if (rmcall)
					clmes_syntax("Expecting FUNCTION <function_name> ROMCALL <romcall_name> [ ARG * ],");
				else
					clmes_syntax("Expecting FUNCTION <function_name> USER [ ARG * ],");
				
				clmes_syntax("but both RETURN or ARG not found.");
				return;
			case 0:
				if (equterm(flptr, "", 6 - rmcall) == -1)
					clmes_syntax("ARG, but no arguments at the end of the line.");
			else
					if (rmcall)
					{
						term(flptr, rcname, 5);
						if ((args = todigit(rcname)) == -1)
						{
							clmes_syntax("Number of arguments (after ARG) is not a number.");
							return;
						}
						*rcname = '\0';
					}
					else
						for (args = 1; equterm(flptr, "", 6 + args) != -1; args++);
				break;
			case -1:
				args = 0;
				break;
		}
		
		// COMPILE IN
#ifdef CREATE_FUNC_DEBUG
		fprintf(stderr, "DEBUG: Creating function %s, type %s, rcname %s, args %d.\n",
			funcname, temptype == function_user ? "USER" : "ROMCALL", rcname, args);
#endif
		totable(temptype, funcname, rcname, args, decline, 0);
	}
}

void newrcall(struct FileLine *flptr, FILE *outfile)
{
	char *rcname, *address = "";
	
#ifdef ROMCALL_DEBUG

        clmes_debug("ROMCALL keyword found.", 2);
#endif

	if (flptr->NumTerms > 2 && flptr->NumTerms != 4)
	{
		clmes_syntax("Expecting ROMCALL <romcall_name> [ = <address> ].");
		return;
	}
		
	if (flptr->NumTerms < 2)
	{
		clmes_syntax("Expecting ROMCALL <romcall_name>, but RETURN at <romcall_name>");
		return;
	}
	
	rcname = flptr->Terms[2];
	
	if (flptr->NumTerms == 4)
		if (!strcmp(flptr->Terms[3], "=") || !strcmp(flptr->Terms[3], "AT"))
		{
			if (todigit(address = flptr->Terms[4]) == -1)
			{
				clmes_syntax("Expecting ROMCALL <romcall_name> = <address>,");
				clmes_syntax("but <address> is no address.");
				return;
			}	
		}	
		else
		{
			clmes_syntax("Expecting ROMCALL <romcall_name> [ AT <address> ].");
			return;
		}
	
	if (checkex(rcname, flptr))
		return;

	if (*address) fprintf(outfile, "%s%c  .equ  %s\n", rcname, \
				tasm ? ' ' : ':', address);

	totable(romcall, rcname, address, 0, line, 0);
}

void newvar(struct FileLine *flptr, FILE *outfile)
{
	int args;
	int vartype;			// 1 = SYSVAR, 2 = STRING, 3 = BYTE
					// 4 = INT, 5 = CONST (see bzc.h)
	char newvar[VAR_LENGTH];
	char andvar[VAR_LENGTH];
	char varstring[VAR_STRING_LENGTH];

	char *adress = "";

	struct SaveTab *newtab;
	
#ifdef NEW_VAR_DEBUG

        clmes_debug("VAR keyword found.", 2);
#endif

	if (flptr->NumTerms == 1)
	{
		clmes_syntax("Expecting <declaration_type> <variable_name>, but RETURN at <variable_name>");
		return;
	}

	term(flptr, newvar, 2);

	if(checkex(newvar, flptr))
		return;

	if (vartype = equterm(flptr, "SYSVAR", 1))

		for (vartype = 1; vartype < 5 && equterm(flptr, KeyWords[++vartype - 2 + SYSTEM_AT], 1););
	else
		vartype = 1;
	 
	// vartype should now contain the variable type.
		
	switch (vartype)
	{
		case 1:
		case 5:
			if (flptr->NumTerms != 2 && flptr->NumTerms != 4)
			{
				if (vartype == 1)
					clmes_syntax("Expecting SYSVAR <variable_name> [ = <address> ].");
				else
					clmes_syntax("Expecting CONST <variable_name> [ = <number> ].");
				return;
			}
			if (flptr->NumTerms == 4)
				if (strcmp(flptr->Terms[3],"=") && strcmp(flptr->Terms[3], "AT"))
				{
					if (vartype == 1)
						clmes_syntax("Expecting SYSVAR <variable_name> [ = <number> ].");
					else
						clmes_syntax("Expecting CONST <variable_name> [ = <number> ].");
					clmes_syntax("but <number> is no number.");
					return;
				}
				else
				{
					if ((args = todigit(adress = flptr->Terms[4])) == -1)
					{
						if (vartype == 1)
							clmes_syntax("Expecting SYSVAR <variable_name> = <address>,");
						else
							clmes_syntax("Expecting CONST <variable_name> = <address>,");
						return;
					}
					fprintf(outfile, "%s%c  .equ  %d\n", flptr->Terms[2], \
						tasm ? ' ' : ':', todigit(flptr->Terms[4]));
				}
			break;
		case 2:		// STRING

			if (flptr->NumTerms != 4)
			{
				clmes_syntax("Expecting STRING <variable_name> = <string>,");
				if (flptr->NumTerms < 4)
					if (flptr->NumTerms == 3)
						clmes_syntax("but RETURN at <string>.");
					else
						clmes_syntax("but RETURN at '='.");
				else
					clmes_syntax("but no RETURN after <string>.");
				return;
			}
			
			if (equterm(flptr, "=", 3))
			{
				clmes_syntax("Expecting STRING <variable_name> = <string>,");
				clmes_syntax("but no '=' at keyword 3.");
				return;
			}

			args = 4;
			break;
			
		case 3:		// BYTE or
		case 4:		// INT

			if ((args = equterm(flptr, "=", 3)) == 1)
			{
				clmes_syntax("Expecting <declaration_type> <variable_name> [ = <number> ],");
				clmes_syntax("but no RETURN and no '=' at keyword 3.");
				return;
			}

			if (args)
				args = 0;
			else
			{
				term(flptr, varstring, 4);
				if (*varstring && *varstring != '\n')
					if ((args = todigit(varstring)) == -1)
					{
						if (vartype == 3)
							clmes_syntax("Expecting BYTE <variable_name> = <number>,");
						else
							clmes_syntax("Expecting INT <variable_name> = <number>,");

						clmes_syntax("but <number> is not a number.");
						clmes_syntax("<number> may NOT have a sign (must allways be positive).");
						return;
					}
				break;
			}
	}

	// Let's put it into the table

	newtab = totable(var, newvar, adress, args, line, vartype);
	if (vartype == 1 || vartype == 3 || vartype == 4)
	{
		addand(andvar, newvar);
		if (vartype == 1)
			totable(var, andvar, adress, args, line, 5);
		else
			totable(var, andvar, newtab->Name, 0, line, 5);
	}
}

void newsprite(struct FileLine *fline)
{
	int i, x, y;
	int width, height;
	int black[MAX_SPRITE_LINES], white[MAX_SPRITE_LINES];
	int *words_array;
	int *words_mem;
	char c;
	char newsprite[VAR_LENGTH];
	FILE *spritefile;

#ifdef SPRITE_DEBUG

        clmes_debug("SPRITE keyword found.", 2);
#endif
	if (fline->NumTerms != 4 || strcmp(fline->Terms[3], "FROM"))
	{
		clmes_syntax("Wrong (number of) arguments, should be:");
		clmes_syntax("SPRITE <sprite_name> FROM <sprite_input_file>");
		return;
	}
		
	term(fline, newsprite, 2);

	if (checkex(newsprite, fline))
		return;

	if (!(spritefile = fopen(fline->Terms[4], "r")))
	{
		clmes_syntax("Can't open SPRITE-file.");
		exit_io();
	}
		
	// Start reading the file to find the x and y of the sprite:
	
	i = 0;
	x = 0;
	height = 0;
	
	for (c = getc(spritefile); c != EOF && i != -1; c = getc(spritefile))
	{
		switch (c)
		{
			case '1':
			case '0':
			case ' ':
				i++;
			case '\r':
				break;

			case '\n':
				if (i > x)
					x = i;
				i = 0;
				height++;
				break;
			default:
				clmes_syntax("Wrong characters in SPRITE file,");
				clmes_syntax("Build your sprite with 1's, 0's and SPACES.");
				i = -1;
				break;
		}
	}

	if (i)
		height++;
	
	if (!x)
	{
		clmes_syntax("SPRITE has a width of 0.");
		i = -1;
	}
	
	if (i == -1)
	{
		fclose(spritefile);
		return;
	}
	
	// Calculate the sprite demensions:
	
	width = x / 8;
	if (x % 8) width++;

	fseek(spritefile, 0, SEEK_SET);

	x = 0;
	
	// Initialise the arrays and fill them:
	
	for (i = 0; i < MAX_SPRITE_LINES; i++)
	{
		white[i] = 0;
		black[i] = 0;
	}	

	y = 0;
	
	for (c = getc(spritefile); c != EOF; c = getc(spritefile))
	{
		i = poweroftwo(7 - x % 8);
		
		switch (c)
		{
			case '0':
				white[x / 8] = white[x / 8] + i;
				x++;
				break;
			case '1':
				black[x / 8] = black[x / 8] + i;
				x++;
				break;
			case '\n':
				x = 8 * width * ++y;
				break;				
		}
		if (width * y >= MAX_SPRITE_LINES)
			exit_LowSprite();
	}
	
	fclose(spritefile);

	// Put everything in the array:
	
	if (!(words_mem = words_array = (int *) getmem(sizeof(int) * width * height)))
		exit_nomem();
	
	*words_array++ = width * 256 + height; // Some sort of 'index'
	
	for (i = 1; i < height * width;)
	{
		x = i++;
		*words_array++ = black[x] * 256 + white[x];
	//	fprintf(stderr, "%d\n", *(words_array - 1));
	}

	// Put the SPRITE to the table.
	
	totable(array, fline->Terms[2], "", width * height, line, ptrtoint(words_mem));
}

void addand(char *newvarname, char *oldvarname)
{
	*newvarname++ = '&';
	while (*newvarname++ = *oldvarname++);
}
