/***************************************************************************
 *   Copyright (C) 2005 by Erik van 't Wout                                *
 *   bugs@hofhom.nl                                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bzc.h"

struct SaveTab *totable(enum SaveType type, char *tempname, char *realname, int args, int atline, int lastline)
{
	char *name;
	struct SaveTab *table;

	if (!*curfunc || globalvar || \
		((type != var || lastline == 1 || lastline == 5) && type != array))
	{
		name = tempname;
		globalvar = 0;
	}
	else
	{
		name = getmem(VAR_LENGTH);
		strcpy(name, curfunc);
		strcat(name, "_");
		strcat(name, tempname);
	}
	
	if (!(table = findtable(name, 0)))
	{
		table = addtable();
		strcpy(table->Name, name);
	}
	
        if (type != -1)		table->Type = type;
	if (args != -1)		table->Args = args;
	if (atline != -1)	table->Line = atline;
	if (lastline != -1)	table->Lastline = lastline;
	if (*realname != '~')	strcpy(table->Realname, realname);
	
#ifdef TABLE_DEBUG
	fprintf(stderr, "DEBUG: New [%d] Table %s (= %s), args: %d (line %d - %d)\n",
			(int) table->Type, table->Name, table->Realname, table->Args,
			table->Line, table->Lastline);
#endif
	return(table);
}

int usetable(char *name)
{
	struct SaveTab *table;
	
	if (!(table = findtable(name, 0)))
		return(-1);
	
	if (table->Used != -1) table->Used++;
	return(0);
}

struct SaveTab *addtable() 
{
	struct SaveTab *curtab = tabptr;
	struct SaveTab *oldtab = curtab;

	if (curtab)
		for(oldtab = curtab; curtab = curtab->Next; oldtab = curtab);
	
	if (!(curtab = (struct SaveTab * ) getmem(sizeof(struct SaveTab))))
#ifdef TABLE_DEBUG		
	{
		clmes_mem_dp("error while trying to get %d bytes for a new table.", \
							sizeof(struct SaveTab), "");
#endif
		exit_nomem();
#ifdef TABLE_DEBUG
	}
	else
	{
		clmes_mem_dp("allocated %d bytes at %p for a new table.", \
					sizeof(struct SaveTab), curtab);
	}
#endif
	if(tabptr)
		oldtab->Next = curtab;
	else
		tabptr = curtab; 

	curtab->Line = 0;		//
	curtab->Lastline = 0;		//
	curtab->Used = 0;		//
	curtab->Args = 0;		// Just for the sake of tidyness,
	*curtab->Name = '\0';		// initialise everything to zero's.
	*curtab->Realname = '\0';	//
	curtab->Type = (enum SaveType) 0;
	curtab->Next = NULL;		//
		
	return(curtab);
}

int ffirstline(char *funcname)
{
	struct SaveTab *tabptr;

	if (!(tabptr = findtable(funcname, 0)))
		return(1);

	if (tabptr->Type != function_user)
		return(2);

	tabptr->Line = line;
	CurFunc = tabptr;
	comflag = Code;
	return(0);
}

void flastline()
{
	CurFunc->Lastline = line;
	comflag = Noflag;
}


struct SaveTab *findtable(char *tempname, int notglobal)
{
	char name[VAR_LENGTH];
	char seekname[VAR_LENGTH];
	char *fullname;
	int i;
	struct SaveTab *seekptr;

	transtoup(name, tempname);
	
	if (*curfunc)
	{
		fullname = getmem(VAR_LENGTH);
		transtoup(fullname, curfunc);
		strcat(fullname, "_");
		strcat(fullname, name);
	}
	else
		fullname = name;
	
	for (seekptr = tabptr; seekptr; seekptr = seekptr->Next)
	{
		transtoup(seekname, seekptr->Name);
		if (!strcmp(seekname, fullname))
			break;
	}

	if (!seekptr && *curfunc && !notglobal)
		for (seekptr = tabptr; seekptr; seekptr = seekptr->Next)
		{
			transtoup(seekname, seekptr->Name);
			if (!strcmp(seekname, name))
				break;
		}

	if (!seekptr && (i = todigit(name)) != -1)
	{
		seekptr = addtable();
		strcpy(seekptr->Name, name);
		totable(var, name, "", i, line, 5);
	}
	
	return(seekptr);
}

void init_reg() 
{
	struct SaveTab *saveTab;

	// We use line number 0 (which doesn't exists),
	// as declare line (the second 0)
	
	totable(reg8, "A", "", 0, 0, 0);
	totable(reg8, "B", "", 0, 0, 0);
	totable(reg8, "C", "", 0, 0, 0);
	totable(reg8, "D", "", 0, 0, 0);
	totable(reg8, "E", "", 0, 0, 0);
	totable(reg8, "F", "", 0, 0, 0);
	totable(reg16, "HL", "", 0, 0, 0);
	totable(reg16, "BC", "", 0, 0, 0);
	totable(reg16, "DE", "", 0, 0, 0);
	totable(reg16, "IX", "", 0, 0, 0);
	totable(reg16, "IY", "", 0, 0, 0);
	
	// Declare a temporarely place for the RET-pointer
	// of the program itself.
		
	saveTab = totable(var, "__RETPointer", "", 0, 0, 4);
	if (!os) saveTab->Used++;
}

void forcevar(struct FileLine *flptr)
{
	struct SaveTab *stab;

	if (flptr->NumTerms != 2)
	{
		clmes_syntax("Expecting FORCE <var>.");
		return;
	}
	if (!(stab = findtable(flptr->Terms[2], 0)))
	{
		clmes_syntax("Expecting FORCE <var>,");
		clmes_syntax("but <var> is no VAR and no FUNCTION.");
		return;
	}
	if (stab->Used > -1) stab->Used++;
	return;
}

char *makeid(char *begin, char *end)
{
	char *tempchar, *c;
	int tempnum;
	int i= 5;

	tempchar = getmem(VAR_LENGTH);
	c = tempchar;

	for (; *begin; *tempchar++ = *begin++)
		if (i++ > VAR_LENGTH)
			exit_lowline();

	*tempchar++ = '_';
	*tempchar++ = '_';
	
	for (tempnum = ++idnum; tempnum;)
		if (i++ > VAR_LENGTH)
			exit_lowline();
		else if (tempnum > 26)
		{
			tempnum -= 26;
			*tempchar++ = 'Z';
		}
		else
		{
			*tempchar++ = (char) (tempnum + 64);
			tempnum = 0;
		}

	*tempchar++ = '_';
	*tempchar++ = '_';
	for (; *end; *tempchar++ = *end++)
		if (i++ > VAR_LENGTH)
			exit_lowline();
	return(c);
}
